
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2001 Konrad Haenel ( alterego1@gmx.at )
 *  Copyright (C) 2000 William W. Wong ( williamw@jps.net )
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex.download;


import java.io.*;
import java.net.*;
import java.util.*;

import phex.*;
import phex.config.*;
import phex.download.*;
import phex.interfaces.*;
import phex.msg.*;
import phex.utils.*;


public class DownloadFile implements Serializable, ISearch
{
	public static final int	sQueued = 1;
	public static final int	sDownloading = 2;
	public static final int	sCompleted = 4;
	public static final int	sError = 5;
	public static final int	sErrorMaxRetry = 12;
	public static final int	sStopping = 7;
	public static final int	sStopped = 8;
	public static final int	sRemoved = 9;
	public static final int	sConnecting = 10;
	public static final int	sRequestPushTransfer = 11;
	public static final int	sRetryWait = 13;


	private String			mFilename;							// Fields to identify the file
	private long			mFileSize;							// Fields to identify the file
	private String			mShortname;							// Filename without the path info for constructing local filename to save.
	private Vector			mRemoteCandidates = new Vector();
	private int				mCurrCandidate;
	private int				mStatus = sQueued;
	private String			mStatusText = "";
	private long			mStatusTime = 0;
	private String			mLog = "";
	private String			mSearchTerm = "";
	private int			mStartSize = 0;
	private long			mTransferredSize = 0;
	private long			mStartingTime = 0;
	private int				mTransferRate = 0;
	private DownloadWorker	mDownloadWorker = null;
	private int				mFileOffset = 0;
	private String			mRemoteAppName = null;
	private DownloadManager	mDownloadMgr;
	private Search			mSearch = null;
	private int				mRetryCount = 0;
	private long			mNextRetryTime = 0;


	private DownloadFile()
	{
		mDownloadMgr = ServiceManager.getDownloadManager();
	}


	public DownloadFile(RemoteFile rfile)
	{
		mShortname = mFilename = rfile.getFilename();
		mFileSize = rfile.getFileSize();
		
		mRemoteCandidates.addElement(rfile);
		mCurrCandidate = 0;

		mDownloadMgr = ServiceManager.getDownloadManager();

		// Some servants just return the path along with the filename.
		// Path info is platform-dependent.  Get rid of it.
		int	i = mFilename.lastIndexOf("/");
		if (i != -1)
		{
			mShortname = mFilename.substring(i + 1);
		}
		else
		{
			i = mFilename.lastIndexOf("\\");
			if (i != -1)
			{
				mShortname = mFilename.substring(i + 1);
			}
		}

		mSearchTerm = mShortname;
	}


	public String getFilename()
	{
		return (mFilename);
	}


	public long getFileSize()
	{
		return mFileSize;
	}


	public int getCurrCandidate()
	{
		return mCurrCandidate;
	}


	public boolean advanceCurrCandidate()
	{
		mCurrCandidate++;
		if (mCurrCandidate >= mRemoteCandidates.size())
		{
			mCurrCandidate = 0;

			// Reach end of candidate list and warp around.
			return true;
		}

		// Not reach end of candidate list yet.
		return false;
	}


	public RemoteFile getCurrentRemoteFile()
	{
		if (mCurrCandidate >= 0)
			return (RemoteFile)mRemoteCandidates.elementAt(mCurrCandidate);
		return null;
	}


/**********************************************************
*      Candidate methods:                                 *
**********************************************************/

	public Vector getRemoteCandidates()
	{
		return mRemoteCandidates;
	}
	
	
	public int addRemoteCandidate(RemoteFile rfile)
	{
		// Check for existing one.  Return it if it exists.
		for (int i = 0; i < mRemoteCandidates.size(); i++)
		{
			RemoteFile	file = (RemoteFile)mRemoteCandidates.elementAt(i);
			if (file.equals(rfile))
				return i;
		}

		// Add as new one.
		mRemoteCandidates.addElement(rfile);

		mDownloadMgr.candidateChanged();
		return mRemoteCandidates.size() - 1;
	}


	public void setCandidateCurrent(int index)
	{
		if (index < 0 || index >= mRemoteCandidates.size() || index == mCurrCandidate)
			return;

		mCurrCandidate = index;
		mDownloadMgr.candidateChanged();
	}


	public void removeRemoteCandidate(int index)
	{
		if (index < 0 || index >= mRemoteCandidates.size() || index == mCurrCandidate)
			return;

		mRemoteCandidates.removeElementAt(index);
		if (mCurrCandidate >= mRemoteCandidates.size())
		{
			mCurrCandidate = mRemoteCandidates.size() - 1;
		}

		mDownloadMgr.candidateChanged();
	}


	public String getSearchTerm()
	{
		return mSearchTerm;
	}


	public void setSearchTerm(String term)
	{
		mSearchTerm = term;
	}
	
	public void searchNewCandidate()
	{
		mCurrCandidate = 0;
		while ( mRemoteCandidates.size() > 1 )
		{
			mRemoteCandidates.remove( 1 );
		}

		if (getIsSearching())
			return;

		MsgManager	mm = ServiceManager.getMsgManager();
		mSearch = mm.sendBackgroundSearch(mSearchTerm, this);
		mDownloadMgr.candidateChanged();
				
		mDownloadMgr.candidateChanged();
	}


	public void searchCandidate()
	{
		if (getIsSearching())
			return;

		MsgManager	mm = ServiceManager.getMsgManager();
		mSearch = mm.sendBackgroundSearch(mSearchTerm, this);
		mDownloadMgr.candidateChanged();
	}


	public void stopSearchCandidate()
	{
		if (!getIsSearching())
			return;

		MsgManager	mm = ServiceManager.getMsgManager();
		mm.stopBackgroundSearch(mSearch);
		mDownloadMgr.candidateChanged();
	}


	public Search getSearchObj()
	{
		return mSearch;
	}
	

/**********************************************************
*      download methods:                                  *
**********************************************************/

	public long getStartingTime()
	{
		return mStartingTime;
	}


	public void setStartingTime(long startingTime)
	{
		mStartingTime = startingTime;
	}


	public String getRangeHeader(int startPos)
	{
		if (startPos == 0)
			return "";

//		return "Range: bytes=" + startPos + "-" (mFileSize - 1) + "\r\n";
		return "Range: bytes=" + startPos + "-\r\n";
	}


	public int getStatus()
	{
		return mStatus;
	}


	public String getStatusName()
	{
		switch (mStatus)
		{
			case sQueued:
				return "queued...";

			case sDownloading:
				return "downloading...";

			case sCompleted:
				return "COMPLETED";

			case sError:
				return  mStatusText;

			case sStopping:
				return "stopping...";

			case sStopped:
				return "STOPPED";

			case sConnecting:
				return "connecting...";
				
			case sRequestPushTransfer:
				return "requesting push transfer...";

			case sErrorMaxRetry:
				return "Error: retried too many times.";
				
			case sRetryWait:
				return "waiting for " + ServiceManager.sCfg.mDownloadRetryWait/1000 + " seconds...";
		}
		return "Error: unknown status";
	}


	public void setStatus(int status)
	{
		setStatus(status, "");
	}


	public void setStatus(int status, String text)
	{
		mStatus = status;
		mStatusText = text;
		mStatusTime = System.currentTimeMillis();
		mDownloadMgr.downloadChanged();
	}


	public boolean isErrorStatusExpired()
	{
		if (mStatus == sError)
		{
			if (System.currentTimeMillis() - mStatusTime > 5000)
			{
				return true;
			}
		}
		return false;
	}


	public boolean getIsSearching()
	{
		if (mSearch == null)
			return false;

		return (mSearch.getSearching());
	}


	public String getRemoteAppName()
	{
		return mRemoteAppName;
	}


	public void setRemoteAppName(String appName)
	{
		mRemoteAppName = appName;
		mDownloadMgr.downloadChanged();
	}


	public void setLog(String log)
	{
		mLog = log;
	}


	public void appendLog(String log)
	{
		if (mLog != null)
			mLog += log + "\n";
		else
			mLog = log + "\n";
	}


	public String getLog()
	{
		return mLog;
	}


	public int getStartSize()
	{
		return mStartSize;
	}


	public void setStartSize(int size)
	{
		mStartSize = size;
	}


	public long getTransferredSize()
	{
		return mTransferredSize;
	}


	public void setTransferredSize(long TransferredSize)
	{
		mTransferredSize = TransferredSize;

		if (mStartingTime != 0)
		{
			long	secs = (System.currentTimeMillis() - mStartingTime) / 1000 + 1;

			mTransferRate = (int)((mTransferredSize - mStartSize) / secs);
		}
		mDownloadMgr.downloadChanged();
	}


	public int getTransferRate()
	{
		return mTransferRate;
	}


	public int getRetryCount()
	{
		return mRetryCount;
	}


	public void setRetryCount(int count)
	{
		mRetryCount = count;
	}


	public long getNextRetryTime()
	{
		return mNextRetryTime;
	}


	public void setNextRetryTime(long nextRetryTime)
	{
		mNextRetryTime = nextRetryTime;
	}


	public DownloadWorker getDownloadWorker()
	{
		return mDownloadWorker;
	}


	public void setDownloadWorker(DownloadWorker worker)
	{
		mDownloadWorker = worker;
	}


	public String getLocalFilename()
	{
		String	downloadDir = ServiceManager.sCfg.mDownloadDir;

		if (downloadDir.endsWith(File.separator))
			downloadDir = downloadDir.substring(0, downloadDir.length() - 1);

		return (downloadDir + File.separator + StrUtil.convertToLocalSystemFilename(mShortname));
	}


	public String getDownloadName()
	{
		// resolve: have shorter suffix because the Mac has a 31 character limit.
		return getLocalFilename() + ".dl";
//		return getLocalFilename() + ".download-" + mFileSize + "-" + mFileOffset;
	}


	public String toString()
	{
		return mFilename + "(" + mFileSize + ")";
	}


	private void writeObject(java.io.ObjectOutputStream oos)
			throws IOException
	{
		// Use latest version to serialize.
		serialize1(oos);
	}


	private void readObject(java.io.ObjectInputStream ois)
			throws IOException, ClassNotFoundException
	{
		mDownloadMgr = ServiceManager.getDownloadManager();
		int	version = ois.readInt();
		switch (version)
		{
			case 1:
				deserialize1(ois);
				break;

			default:
				throw new IOException("Fail to deserialize datafile.  Unknown verison.");
		}
	}


	private void serialize1(ObjectOutputStream oos)
			throws IOException
	{
		// Write version 1
		oos.writeInt(1);

		oos.writeUTF(mFilename);
		oos.writeLong(mFileSize);
		oos.writeUTF(mShortname);
		oos.writeUTF(mSearchTerm);
		oos.writeInt(mCurrCandidate);
		oos.writeObject(mRemoteCandidates);
		oos.writeInt(mStatus);
	}


	private void deserialize1(ObjectInputStream ois)
			throws IOException, ClassNotFoundException
	{
		mFilename = ois.readUTF();
		mFileSize = ois.readLong();
		mShortname = ois.readUTF();
		mSearchTerm = ois.readUTF();
		mCurrCandidate = ois.readInt();
		mRemoteCandidates = (Vector)ois.readObject();
		mStatus = ois.readInt();
		if (mStatus != sCompleted)
		{
			mStatus = sQueued;
		}
	}



	// ISearch
	public void started(Search searchObj)
	{
		mDownloadMgr.candidateChanged();
	}


	public void stopped(Search searchObj)
	{
		mDownloadMgr.candidateChanged();
	}


	public void gotResult(Search searchObj)
	{
		Vector		results = searchObj.getQueryResults();
		RemoteFile	rfile = (RemoteFile)results.elementAt(results.size() - 1);

		if (rfile.getFilename().equals(mFilename) && rfile.getFileSize() == mFileSize)
			addRemoteCandidate(rfile);

		mDownloadMgr.candidateChanged();
	}

}


