
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex.download;


import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;

import phex.*;
import phex.config.*;
import phex.connection.*;
import phex.download.*;
import phex.host.*;
import phex.interfaces.*;
import phex.msg.*;


public class DownloadManager
{
	private Vector				mDownloadList = new Vector();
	private Hashtable			mPendingPushTransfer = new Hashtable();
	private int					mWorkerCount = 0;
	private boolean				mShutdownAllWorkers = false;
	private int					mDownloadingCount = 0;
	private IDownloadChanged	mDownloadChangedListener;
	private IDownloadCandidate	mCandidateListener;
	private DataChanger			mStatusChangedListener = new DataChanger();


	public DownloadManager()
	{
	}


	public void setDownloadChangedListener(IDownloadChanged l)
	{
		mDownloadChangedListener = l;
	}


	public void setCandidateListener(IDownloadCandidate l)
	{
		mCandidateListener = l;
	}


	public void addStatusChangedListener(IDataChangedListener listener)
	{
		mStatusChangedListener.addListener(listener);
	}
	

	public void downloadChanged()
	{
		if (mDownloadChangedListener != null)
			mDownloadChangedListener.downloadChanged();

		mStatusChangedListener.dataChanged(null);
	}


	public void candidateChanged()
	{
		if (mCandidateListener != null)
			mCandidateListener.candidatesHaveChanged();
	}


	public IDownloadCandidate getCandidateListener()
	{
		return mCandidateListener;
	}


	public void init()
	{
		try
		{
			loadDownloadInfo();
		}
		catch (Exception e)
		{
			// ignore
		}
		createEnoughWorkers();
		downloadChanged();
	}


	// Check to see if there are enough workers to perform the maximum allowed
	// number of concurrent downloads.  If not enough, create them.
	public synchronized void createEnoughWorkers()
	{
		// One worker per one concurrent download.
		// Create upto the max concurrent downloads allowed.
		while (mWorkerCount < ServiceManager.sCfg.mMaxDownload)
		{
			new DownloadWorker(this);
			mWorkerCount++;
		}
	}


	synchronized boolean isTooManyWorker()
	{
		if (mWorkerCount > ServiceManager.sCfg.mMaxDownload)
		{
			// The DownloadWorker who calls this must exit now.
			mWorkerCount--;
			return true;
		}
		return false;
	}


	private boolean tooManyDownloadPerIP(DownloadFile download1)
	{
		int			count = 0;
		String		remoteHost = download1.getCurrentRemoteFile().getRemoteHost();


		for (int i = 0; i < mDownloadList.size(); i++)
		{
			DownloadFile	download = (DownloadFile)mDownloadList.elementAt(i);
			if (download == download1)
				continue;

			RemoteFile		file = download.getCurrentRemoteFile();
			if (file.getRemoteHost().equals(remoteHost) &&
				(download.getStatus() == DownloadFile.sDownloading ||
				 download.getStatus() == DownloadFile.sConnecting ||
				 download.getStatus() == DownloadFile.sRequestPushTransfer))
			{
				count++;
			}
		}

		return (count >= ServiceManager.sCfg.mMaxDownloadPerIP);
	}


	public synchronized boolean addFileToDownload(RemoteFile rfile)
	{
		// Check for duplicate filename in the existing files to download.
		for (int i = 0; i < mDownloadList.size(); i++)
		{
			DownloadFile	existingFile = (DownloadFile)mDownloadList.elementAt(i);

			if (existingFile.getFilename().equals(rfile.getFilename()))
			{
				if (existingFile.getFileSize() == rfile.getFileSize())
				{
					existingFile.addRemoteCandidate(rfile);
					try
					{
						saveDownloadInfo();
					}
					catch (Exception e)
					{
						e.printStackTrace();
					}
					notifyAll();
					downloadChanged();
					return true;
				}

				// Duplicate filename being downloaded.
				return false;
			}
		}

		// Make sure there are enough workders to do download.
		createEnoughWorkers();

		DownloadFile		download = new DownloadFile(rfile);
		mDownloadList.addElement(download);
		download.setStatus(DownloadFile.sQueued);
		try
		{
			saveDownloadInfo();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		notifyAll();
		return true;
	}


	public synchronized DownloadFile getNextFileToDownload(DownloadWorker worker)
	{
		while (true)
		{
			if (mShutdownAllWorkers)
				return null;

			// resolve: verify that notifyall when number of max workers changed.
			if (isTooManyWorker())
			{
				// Too many workers to do download.  I should exit the thread.
				// Put the file back into queuing mode.
				return null;
			}

			for (int i = 0; i < mDownloadList.size(); i++)
			{
				DownloadFile	download = (DownloadFile)mDownloadList.elementAt(i);

				if (download.getStatus() == DownloadFile.sQueued)
				{
					if (tooManyDownloadPerIP(download))
						continue;
					
					download.setStatus(DownloadFile.sConnecting);
					download.setDownloadWorker(worker);
					download.setStartingTime(System.currentTimeMillis());
					return download;
				}
			}

			// No download waiting to be processed.  Go to sleep.
			try
			{
				wait();
			}
			catch (Exception e)
			{
				// ignore
			}
		}
	}


	public void saveDownloadInfo()
			throws IOException
	{
		FileOutputStream	fos = new FileOutputStream(ServiceManager.getDownloadSaveFilename());
		ObjectOutputStream	oos = new ObjectOutputStream(fos);

		try
		{
			// Use latest version to serialize.
			serialize1(oos);
		}
		finally
		{
			oos.close();
		}
	}


	public void loadDownloadInfo()
			throws Exception
	{
		FileInputStream		fis = new FileInputStream(ServiceManager.getDownloadSaveFilename());
		ObjectInputStream	ois = new ObjectInputStream(fis);

		try
		{
			int	version = ois.readInt();
			switch (version)
			{
				case 1:
					deserialize1(ois);
					break;

				default:
					throw new IOException("Fail to deserialize datafile.  Unknown verison.");
			}
		}
		catch (java.io.FileNotFoundException e1)
		{
			// Ignore
		}
		finally
		{
			ois.close();
		}
	}


	private void serialize1(ObjectOutputStream oos)
			throws IOException
	{
		// Write version 1
		oos.writeInt(1);

		oos.writeObject(mDownloadList);
	}


	private void deserialize1(ObjectInputStream ois)
			throws IOException, ClassNotFoundException
	{
		mDownloadList = (Vector)ois.readObject();
	}


	public synchronized void incDownloadingCount()
	{
		mDownloadingCount++;
	}


	public synchronized void decDownloadingCount()
	{
		mDownloadingCount--;
	}


	public int getDownloadingCount()
	{
		return mDownloadingCount;
	}


	public synchronized void stopDownload(int index)
	{
		if (index >= mDownloadList.size())
			return;

		DownloadFile	download = (DownloadFile)mDownloadList.elementAt(index);
		DownloadWorker	worker = download.getDownloadWorker();
		if (worker != null)
		{
			// Download in progress.  Tell the worker to stop it.
			download.setStatus(DownloadFile.sStopping);
			worker.stopDownload();
		}
		else
		{
			// Just change the status to sStopped.
			download.setStatus(DownloadFile.sStopped);
		}
	}


	public synchronized void resumeDownload(int index)
	{
		if (index >= mDownloadList.size())
			return;

		DownloadFile	download = (DownloadFile)mDownloadList.elementAt(index);
		if (download.getStatus() == DownloadFile.sStopped ||
			download.getStatus() == DownloadFile.sError)
		{
			download.setStatus(DownloadFile.sQueued);
			download.setRetryCount(1);
			notifyAll();
		}
	}


	public synchronized void removeDownload(int index)
	{
		if (index >= mDownloadList.size())
			return;

		DownloadFile	download = (DownloadFile)mDownloadList.elementAt(index);
		mDownloadList.removeElementAt(index);

		DownloadWorker	worker = download.getDownloadWorker();
		if (worker != null)
		{
			// Download in progress.  Tell the worker to stop it and remove file.
			download.setStatus(DownloadFile.sRemoved);
			worker.stopDownload();
		}
		else
		{
			// Remove intermediate file.
			File		f = new File(download.getDownloadName());
			if (f.exists())
			{
				f.delete();
			}
		}

		try
		{
			saveDownloadInfo();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		downloadChanged();
		candidateChanged();
	}


	public synchronized void removeCompleted()
	{
		for (int i = 0; i < mDownloadList.size(); )
		{
			DownloadFile	download = (DownloadFile)mDownloadList.elementAt(i);

			if (download.getStatus() == DownloadFile.sCompleted)
			{
				mDownloadList.removeElementAt(i);
			}
			else
			{
				i++;
			}
		}
		downloadChanged();
	}


	public Vector getDownloadList()
	{
		return mDownloadList;
	}


	public int getDownloadCount()
	{
		return mDownloadList.size();
	}


	// Called by ReadWorker to read a file being pushed from the remote host.
	public synchronized boolean notifyReadPushedFile(ConnectionRequest req, Host remoteHost)
	{
		// GIV <file-ref-num>:<ClientID GUID in hexdec>/<filename>\n\n
		String	key = req.getHeader("FileRef") + ":" + req.getHeader("ClientID") + "/" + req.getHeader("Filename");
		if (mPendingPushTransfer.get(key) != null)
		{
			// Transfer remoteHost and its connection to the download thread.
			mPendingPushTransfer.put(key, remoteHost);

			// Wait up the download thread.
			notifyAll();
			return true;
		}

		return false;
	}


	// Called by DownloadWorker to send out a PushRequest and wait for incoming push connection.
	public synchronized Host pushRequestTransfer(Host oldHost, DownloadFile download)
			throws Exception
	{
		download.setStatus(DownloadFile.sRequestPushTransfer);

		RemoteFile		file = download.getCurrentRemoteFile();

		// Prepare the PushRequest msg.
		MsgPushRequest	msg = new MsgPushRequest(new MsgHeader());
		msg.setClientID(file.getRemoteClientID());
		msg.setFileIndex(file.getFileIndex());
		msg.setRequestPort((short)ServiceManager.getListener().getListeningPort());
		msg.setRequestHost(ServiceManager.getListener().getSelfHostAddress());

		// Save it in pending transfer table using greeting as the key.
		// GIV <file-ref-num>:<ClientID GUID in hexdec>/<filename>\n\n
		String			key = msg.getFileIndex() + ":" +
							  msg.getClientID().toHexString() + "/" +
							  file.getFilename();
		mPendingPushTransfer.put(key, oldHost);

		download.appendLog(key);

		// Route the PushRequest msg.
		Host			returnHost = ServiceManager.getMsgManager().getPushRouting(msg.getClientID());
		if (returnHost == null)
		{
			throw new Exception("Routing table has no info to route back to sharing host.");
		}
		ServiceManager.getSendManager().queueMsgToSend(returnHost, msg, false);

		download.appendLog("Wait for connection from sharing host");

		// Sleep and wait for the incoming connection from the pushing host.
		long			then = System.currentTimeMillis();
		while (mPendingPushTransfer.get(key) == oldHost)
		{
			long		timeElapsed = System.currentTimeMillis() - then;
			long		timeToSleep = ServiceManager.sCfg.mPushTransferTimeout - timeElapsed;

			if (timeToSleep < 0)
			{
				// Timeout.
				mPendingPushTransfer.remove(key);
				download.appendLog("Time out waiting for connection from sharing host");
				return null;
			}

			try
			{
				wait(timeToSleep + 1000);
			}
			catch (Exception e)
			{
				// ignore
			}
		}

		download.appendLog("Sharing host made connection to me");
		return (Host)mPendingPushTransfer.remove(key);
	}


	// Called by JobsRunner periodically to retry failed download.
	public void retryDownloads()
	{
		int		changedCount = 0;


		for (int i = 0; i < mDownloadList.size(); i++)
		{
			DownloadFile	download = (DownloadFile)mDownloadList.elementAt(i);
			if (download.getStatus() == DownloadFile.sError)
			{
				if (download.isErrorStatusExpired())
				{
					changedCount++;
					if (download.getRetryCount() > ServiceManager.sCfg.mDownloadMaxRetry)
					{
						download.setStatus(DownloadFile.sErrorMaxRetry);
						continue;
					}

					if (download.advanceCurrCandidate())
					{
						// Have re-tried on all candidates.
						// Sleep to wait for next retry.
						download.setStatus(DownloadFile.sRetryWait);
						download.setNextRetryTime(System.currentTimeMillis() + ServiceManager.sCfg.mDownloadRetryWait);
					}
					else
					{
						// Try next candidate.
						download.setRetryCount(download.getRetryCount() + 1);
						download.setStatus(DownloadFile.sQueued);
					}
				}
			}
			else if (download.getStatus() == DownloadFile.sRetryWait)
			{
				if (System.currentTimeMillis() > download.getNextRetryTime())
				{
					changedCount++;
					download.setRetryCount(download.getRetryCount() + 1);
					download.setStatus(DownloadFile.sQueued);
				}
			}
		}

		if (changedCount > 0)
		{
			synchronized (this)
			{
				notifyAll();
			}
			downloadChanged();
			candidateChanged();
		}
	}


	// Called by JobsRunner to do searchCandidate.
	public void searchCandidates()
	{
		for (int i = 0; i < mDownloadList.size(); i++)
		{
			DownloadFile	download = (DownloadFile)mDownloadList.elementAt(i);
			download.searchCandidate();
		}
	}


}


