
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex.host;


import java.util.*;
import java.io.*;

import phex.*;
import phex.config.*;
import phex.interfaces.*;
import phex.msg.*;
import phex.utils.*;


public class HostManager
{
	private Vector					mHosts;
	private Vector					mConnectedHosts;
	private IHostChanged			mHostChangedListener = null;
	private IHostCaughtChanged		mHostCaughtChangedListener = null;
	private DataChanger				mHostCountChangedListener = new DataChanger();
	private DataChanger				mStatusChangedListener = new DataChanger();
	private int						mStatHosts = 0;
	private int						mStatFiles = 0;
	private long					mStatSize = 0;
	private String					mStatSizeStr = "0";
	private int						mStatTakenCount = 0;
	private int						mStatMsgCount = 0;
	private int						mStatDropCount = 0;
	private int						mStatUploadCount = 0;
	private int						mStatDownloadCount = 0;
	private Vector					mHostsCaught = new Vector();
	private Hashtable				mHostsCaughtUnique = new Hashtable();
	private boolean					mHostsCaughtChanged = false;
	private long					mLastSaved = System.currentTimeMillis();
	private long					mRateTime0 = System.currentTimeMillis();
	private int						mRateBytes = 0;
	private int						mOldRate = 0;


	public HostManager()
	{
		mHosts = new Vector();

		mConnectedHosts = new Vector();
		newRateSamplePeriod();
	}


	private void addStarterCacheProxy()
	{
		if (ServiceManager.sCfg.mCurrentNetwork.equals(ServiceManager.sCfg.sGeneralNetwork))
		{
			String[]	list = getStarterHosts();

			for (int i = 0; i < list.length; i++)
			{
				addHostCaught(list[i]);
			}
		}
	}


	public void removeStartHosts()
	{
		String[]		list = getStarterHosts();

		for (int i = 0; i < list.length; i++)
		{
			removeHostCaught(list[i]);
		}
	}


	private String[] getStarterHosts()
	{
		String[]		list = new String[2];

		list[0] = "gnutellahosts.com:6346";
		list[1] = "router.limewire.com:6346";
		return list;
	}


	public Vector getHosts()
	{
		return mHosts;
	}


	public synchronized void connectAllOutgoingHosts()
	{
		for (int i = 0; i < mHosts.size(); i++)
		{
			connectHost(i);
		}
	}


	public synchronized void connectHost(int index)
	{
		if (index >= mHosts.size())
			return;

		Host	host = (Host)mHosts.elementAt(index);

		if (host.getType() != Host.sTypeOutgoing)
			return;

		if (host.getSock() != null)
		{
			return;
		}

		new ReadWorker(host, 0);
	}


	public synchronized void disconnectHost(int index)
	{
		if (index >= mHosts.size())
			return;

		Host	host = (Host)mHosts.elementAt(index);
		host.closeConnection();
	}


	public synchronized void disconnectAll()
	{
		for (int i = 0; i < mHosts.size(); i++)
		{
			disconnectHost(i);
		}
	}


	public synchronized void timeoutOutgoingHosts()
	{
		for (int i = 0; i < mHosts.size(); i++)
		{
			Host	host = (Host)mHosts.elementAt(i);

			if (host.isConnectingTimeout())
			{
				host.setStatus(Host.sStatusTimeout, "");
			}
		}
	}


	public synchronized int addOutgoingHost(String hostAddrStr)
	{
		// Check for duplicate.
		for (int i = 0; i < mHosts.size(); i++)
		{
			Host	host = (Host)mHosts.elementAt(i);

			if (host.getType() == Host.sTypeOutgoing &&
				host.getHostAddr().equals(hostAddrStr))
			{
				return -1;
			}
		}

		int		index = mHosts.size();
		Host	host = new Host(this);
		host.setType(Host.sTypeOutgoing);
		host.setHostAddr(hostAddrStr);
		mHosts.addElement(host);

		// Insert in the front so that will use it first in restart.
		insertHostCaught(hostAddrStr);

		try
		{
			saveHosts();
		}
		catch (IOException e)
		{
//			System.out.println(e);
		}

		return index;
	}


	public synchronized void addIncomingHost(Host host)
	{
		mHosts.addElement(host);
		getHostChangedListener().hostChanged();
	}


	public synchronized void removeHost(int index)
	{
		if (index >= mHosts.size())
			return;

		Host	host = (Host)mHosts.elementAt(index);

		host.closeConnection();
		mHosts.removeElementAt(index);
//		moveToEndOfHostCaught(host.getHostAddr());
	}


	public synchronized void removeAll()
	{
		while (mHosts.size() > 0)
		{
			removeHost(0);
		}
	}


	public int getConnectedHostCount()
	{
		return mConnectedHosts.size();
	}


	public synchronized int cleanupAndGetLiveHostCount()
	{
		int	count = 0;
		int	size = mHosts.size();

		for (int i = 0; i < size; i++)
		{
			Host	host = (Host)mHosts.elementAt(i);
			int		status = host.getStatus();

			if (status == Host.sStatusConnected ||
				status == Host.sStatusConnecting ||
				status == Host.sStatusAccepting)
			{
				count++;
			}
			else
			{
				if (ServiceManager.sCfg.mAutoCleanup && host.isErrorStatusExpired())
				{
					removeHost(i);
					size--;
					i--;
				}
			}
		}
		return count;
	}


	public synchronized Host getConnectedHost(int i)
	{
		return (Host)mConnectedHosts.elementAt(i);
	}


	public synchronized void addConnectedHost(Host host)
	{
		mConnectedHosts.addElement(host);
		setHostCaughtConnectionFailed(host.getHostAddr(), false);
		mHostCountChangedListener.dataChanged(mConnectedHosts);
		mStatusChangedListener.dataChanged(null);
	}


	public synchronized void removeConnectedHost(Host host)
	{
		mConnectedHosts.removeElement(host);
		mHostCountChangedListener.dataChanged(mConnectedHosts);
		mStatusChangedListener.dataChanged(null);
	}


	// Forward broadcast messages to neighbors.
	public synchronized void forwardMsg(IMsg msg, Host fromHost)
	{
		if (decTTL(msg))
		{
//			fromHost.log("TTL expired " + msg);
			return;
		}

		SendManager		sendMgr = ServiceManager.getSendManager();
		Host			remoteHost;

		for (int i = 0; i < mConnectedHosts.size(); i++)
		{
			if ((remoteHost = (Host)mConnectedHosts.elementAt(i)) == fromHost)
			{
				continue;
			}

			if (remoteHost.sendQueueInRed() || remoteHost.latencyInRed())
			{
				// Neighbor is too slow.  Deprecate the neighbor.
				// Don't send broadcast message to neighbor.
				continue;
			}

			// Note that the msg is shared among all sending hosts.
			sendMgr.queueMsgToSend(remoteHost, msg, false);
		}
	}


	public synchronized void sendMsgToHosts(IMsg msg)
	{
		SendManager		sendMgr = ServiceManager.getSendManager();

		for (int i = 0; i < mConnectedHosts.size(); i++)
		{
			// Note that the msg is shared among all sending hosts.
			sendMgr.queueMsgToSend((Host)mConnectedHosts.elementAt(i), msg, false);
		}
	}


	public boolean decTTL(IMsg msg)
	{
		MsgHeader	header = msg.getHeader();
		int			ttl = header.getTTL() - 1;
		
		if (ttl <= 0)
		{
			// Expired
			return true;
		}

		if (ttl > ServiceManager.sCfg.mNetMaxTTL)
		{
			ttl = ServiceManager.sCfg.mNetMaxTTL;
		}

		header.setTTL(ttl);
		header.setHopsTaken(header.getHopsTaken() + 1);

		return false;
	}


	public synchronized void pingNeighbors()
	{
		for (int i = 0; i < mConnectedHosts.size(); i++)
		{
			((Host)mConnectedHosts.elementAt(i)).pingHost();
		}
	}


	public synchronized void checkConnections()
	{
		for (int i = 0; i < mConnectedHosts.size(); i++)
		{
			Host	host = (Host)mConnectedHosts.elementAt(i);

			if (host.tooManyDropPackets())
			{
				host.closeConnection();
				continue;
			}
			if (host.tooMuchLatency())
			{
				host.closeConnection();
				continue;
			}
			if (host.sendQueueTooLong())
			{
				host.closeConnection();
				continue;
			}
		}
	}


	public void loadHosts()
			throws IOException
	{
		File				file = new File(ServiceManager.getHostsFilename());
		BufferedReader		br = new BufferedReader(new FileReader(file));
		String				line;

		while ((line = br.readLine()) != null)
		{
			if (line.startsWith("#"))
				continue;

			// Resolve: Handle old host file.  Can be removed after couple versions.
			StringBuffer	buf = new StringBuffer();
			for (int i = 0; i < line.length(); i++)
			{
				if (line.charAt(i) == '\\' || line.charAt(i) == '=')
				{
					continue;
				}
				buf.append(line.charAt(i));
			}

			line = buf.toString();
			addHostCaught(line);
		}

		br.close();

		addStarterCacheProxy();
	}


	public synchronized void saveHosts()
			throws IOException
	{
		File				file = new File(ServiceManager.getHostsFilename());
		BufferedWriter		bw = new BufferedWriter(new FileWriter(file));

		for (int i = 0; i < mHostsCaught.size(); i++)
		{
			String			host = (String)mHostsCaught.elementAt(i);
			bw.write(host);
			bw.newLine();
		}
		bw.close();
		mHostsCaughtChanged = false;
		mLastSaved = System.currentTimeMillis();
	}


	public synchronized void autoSaveHosts()
	{
		if (mHostsCaughtChanged)
		{
			// Wait for at least 30 seconds before saving.
			if ((System.currentTimeMillis() - mLastSaved) > 30000)
			{
				try
				{
					saveHosts();
				}
				catch (Exception e)
				{
					System.out.println(e);
				}
			}
		}
	}


	public void setHostChangedListener(IHostChanged listener)
	{
		mHostChangedListener = listener;
	}


	public void setHostCaughtChangedListener(IHostCaughtChanged listener)
	{
		mHostCaughtChangedListener = listener;
	}


	public void addHostCountChangedListener(IDataChangedListener listener)
	{
		mHostCountChangedListener.addListener(listener);
	}
	

	public void addStatusChangedListener(IDataChangedListener listener)
	{
		mStatusChangedListener.addListener(listener);
	}
	


	public IHostChanged getHostChangedListener()
	{
		return mHostChangedListener;
	}


	public int getStatHosts()
	{
		return mStatHosts;
	}


	public void setStatHosts(int val)
	{
		mStatHosts = val;
	}


	public synchronized void incStatHosts(int amount)
	{
		mStatHosts += amount;
	}


	public int getStatFiles()
	{
		return mStatFiles;
	}


	public void setStatFiles(int val)
	{
		mStatFiles = val;
	}


	public synchronized void incStatFiles(int amount)
	{
		if (amount < 0)
			return;
		mStatFiles += amount;
	}


	public long getStatSize()
	{
		return mStatSize;
	}


	public void setStatSize(long val)
	{
		mStatSize = val;
		updateStatSizeStr();
	}


	public synchronized void incStatSize(long amount)
	{
		// Sounds bogos.  Ignore
		if (amount < 0)
			return;
		
		mStatSize += amount;
		updateStatSizeStr();
	}


	public String getStatSizeStr()
	{
		return mStatSizeStr;
	}


	private void updateStatSizeStr()
	{
		mStatSizeStr = StrUtil.formatSizeBytes(mStatSize << 10);
	}


	public int getStatTakenCount()
	{
		return mStatTakenCount;
	}


	public void setStatTakenCount(int val)
	{
		mStatTakenCount = val;
	}


	public synchronized void incStatTakenCount(int amount)
	{
		mStatTakenCount += amount;
	}

	
	public int getStatMsgCount()
	{
		return mStatMsgCount;
	}


	public void setStatMsgCount(int val)
	{
		mStatMsgCount = val;
	}


	public synchronized void incStatMsgCount(int amount)
	{
		mStatMsgCount += amount;
	}

	
	public int getStatDropCount()
	{
		return mStatDropCount;
	}


	public void setStatDropCount(int val)
	{
		mStatDropCount = val;
	}


	public synchronized void incStatDropCount(int amount)
	{
		mStatDropCount += amount;
	}

	
	public int getStatUploadCount()
	{
		return mStatUploadCount;
	}


	public void setStatUploadCount(int val)
	{
		mStatUploadCount = val;
	}


	public synchronized void incStatUploadCount(int amount)
	{
		mStatUploadCount += amount;
	}

	
	public int getStatDownloadCount()
	{
		return mStatDownloadCount;
	}


	public void setStatDownloadCount(int val)
	{
		mStatDownloadCount = val;
		mStatusChangedListener.dataChanged(null);
	}


	public synchronized void incStatDownloadCount(int amount)
	{
		mStatDownloadCount += amount;
		mStatusChangedListener.dataChanged(null);
	}


	public void resetStat()
	{
		mStatHosts = 0;
		mStatFiles = 0;
		mStatSize = 0;
		mStatSizeStr = "0";
		mStatTakenCount = 0;
		mStatMsgCount = 0;
		mStatDropCount = 0;
		mStatUploadCount = 0;
		mStatDownloadCount = 0;
		newRateSamplePeriod();
	}


	public void incBytesCount(int count)
	{
		mRateBytes += count;
	}


	public final int getRate()
	{
		if (mRateBytes == 0)
		{
			// No date in new sampling period.  Use old rate for continuation.
			return mOldRate;
		}
		return mRateBytes * 1000 / (int)(System.currentTimeMillis() - mRateTime0 + 1);
	}


	public void newRateSamplePeriod()
	{
		mOldRate = getRate();
		mRateBytes = 0;
		mRateTime0 = System.currentTimeMillis();
	}


	public Vector getHostsCaught()
	{
		return mHostsCaught;
	}


	public synchronized void insertHostCaught(String host)
	{
		if (mHostsCaughtUnique.get(host) == null)
		{
			mHostsCaught.insertElementAt(host, 0);
			mHostsCaughtUnique.put(host, new Boolean(false));
			if (mHostCaughtChangedListener != null)
				mHostCaughtChangedListener.hostCaughtChanged();
		}
	}


	public synchronized boolean addHostCaught(String host)
	{
		if (mHostsCaught.size() > ServiceManager.sCfg.mNetMaxHostToCatch)
		{
			boolean		roomMade = false;
			
			for (int i = 0; i < mHostsCaught.size(); i++)
			{
				String	host2 = (String)mHostsCaught.elementAt(i);
				Boolean	failed = (Boolean)mHostsCaughtUnique.get(host2);
				if (failed != null && failed.booleanValue())
				{
					removeHostCaught(host2);
					roomMade = true;
					break;
				}
			}

			if (!roomMade)
				return false;
		}

		addHostCaughtForced(host);
		return true;
	}


	public synchronized void addHostCaughtForced(String host)
	{
		if (mHostsCaughtUnique.get(host) == null)
		{
			mHostsCaught.addElement(host);
			mHostsCaughtUnique.put(host, new Boolean(false));
			if (mHostCaughtChangedListener != null)
				mHostCaughtChangedListener.hostCaughtChanged();
		}
	}


	public synchronized void removeHostCaught(String host)
	{
		mHostsCaught.removeElement(host);
		mHostsCaughtUnique.remove(host);
		if (mHostCaughtChangedListener != null)
			mHostCaughtChangedListener.hostCaughtChanged();
	}


	public synchronized boolean addAndSaveHostCaught(String host)
	{
		if (addHostCaught(host) == false)
			return false;

		mHostsCaughtChanged = true;

		return true;
	}


	public void setHostCaughtConnectionFailed(String host, boolean failed)
	{
		mHostsCaughtUnique.put(host, new Boolean(failed));
	}


	public synchronized void moveToEndOfHostCaught(String host)
	{
		for (int i = 0; i < mHostsCaught.size(); i++)
		{
			if (((String)mHostsCaught.elementAt(i)).equals(host))
			{
				mHostsCaught.removeElementAt(i);
				mHostsCaught.addElement(host);
				mHostCaughtChangedListener.hostCaughtChanged();
				break;
			}
		}
	}


	public synchronized void resetHostsCaught()
	{
		mHostsCaught.removeAllElements();
		mHostsCaughtUnique.clear();

		addStarterCacheProxy();
		mHostCaughtChangedListener.hostCaughtChanged();
	}


	public int throttleControl(int currSleep, int maxSleep, int currRate, int maxRate)
	{
		if (currRate > maxRate)
		{
			// Sleep more to slow down reading in new message.
			currSleep = currSleep * 2;
			if (currSleep > maxSleep)
				currSleep = maxSleep;
		}
		else if (currRate < maxRate)
		{
			// Sleep less to speed up reading in new message.
			currSleep = (int)((float)currSleep * 2 / 3);
			if (currSleep == 0)
				currSleep = 1;		// avoid 0; otherwise multiplication won't work.
		}

		try
		{
			Thread.sleep(currSleep);
		}
		catch (Exception e)
		{
		}

		return currSleep;
	}



	public boolean isHostIgnored(String ip)
	{
		return isIpInHosts(ip, ServiceManager.sCfg.mNetIgnoredHosts);
	}


	public boolean isHostInvalid(String ip)
	{
		return isIpInHosts(ip, ServiceManager.sCfg.mNetInvalidHosts);
	}


	public boolean isHostFiltered(String ip)
	{
		return (ServiceManager.sCfg.mApplyFilterdHosts &&
				isIpInHosts(ip, ServiceManager.sCfg.mFilteredSearchHosts));
	}


	public boolean isIpInHosts(String ip, Vector hosts)
	{
		int			size = hosts.size();
		String[]	parts = Cfg.ip2parts(ip);
		int			j;

		for (int i = 0; i < size; i++)
		{
			String[]	partmask = (String[])hosts.elementAt(i);
			for (j = 0; j < 4; j++)
			{
				if (partmask[j].equals("*"))
					continue;
				if (!partmask[j].equals(parts[j]))
					break;
			}
			if (j == 4)
				return true;
		}

		return false;
	}


}



