
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex;


import java.io.*;
import java.net.*;
import java.util.*;

import phex.config.*;
import phex.download.*;
import phex.host.*;
import phex.msg.*;


public class JobsRunner implements Runnable
{
	private boolean				mRequestToDie = false;
	private boolean				mRunning = false;
	private boolean				mShutdownCompleted = false;
	private HostManager			mHostMgr = ServiceManager.getHostManager();
	private NetworkManager		mNetworkMgr = ServiceManager.getNetworkManager();
	private MsgManager			mMsgManager = ServiceManager.getMsgManager();
	private DownloadManager		mDownloadMgr = ServiceManager.getDownloadManager();
	private ShareManager		mShareMgr = ServiceManager.getShareManager();
	private boolean				mInitScan = false;
	private boolean				mInitSearch = false;



	JobsRunner()
	{
	}


	public synchronized void startup()
	{
		if (mRunning)
			return;

		ServiceManager.log("Background jobs runner starting...");

		mRequestToDie = false;
		mRunning = true;

		(new Thread(this)).start();
	}


	public boolean getRunning()
	{
		return mRunning;
	}


	public synchronized void shutdown(boolean waitForCompleted)
	{
		// Already dead or been requested to die.
		if (mRequestToDie)
			return;

		ServiceManager.log("Background jobs runner shutting down...");

		// Set flag to die.
		mShutdownCompleted = false;
		mRequestToDie = true;

		if (waitForCompleted)
		{
			// Wait until the thread is dead.
			while (!mShutdownCompleted)
			{
				try
				{
					wait();
				}
				catch (InterruptedException e)
				{
					break;
				}
			}
		}
	}


	public void run()
	{
		ServiceManager.log("Background jobs runner started.");

		MainFrame	mainFrame = ServiceManager.getManager().getMainFrame();
		int			sleepAmount = 1000;
		int			msgRateSampleDuration = 10000 / sleepAmount;
		int			msgRateSampleCount = 0;
		int			initSearchDuration = 60000 / sleepAmount;
		int			counter = 0;

		while (true)
		{
			try
			{
				// Sleep for a while.
				try
				{
					Thread.sleep(sleepAmount);
				}
				catch (Exception e)
				{
				}

				// See if I has been asked to die.
				if (mRequestToDie)
				{
					break;
				}

				// Do the routine jobs.

				msgRateSampleCount++;
				if (msgRateSampleCount > msgRateSampleDuration)
				{
					msgRateSampleCount = 0;
					mHostMgr.newRateSamplePeriod();
				}

				mHostMgr.pingNeighbors();
				mHostMgr.timeoutOutgoingHosts();

				if (counter % 5 == 0)
				{
					mHostMgr.autoSaveHosts();
					mainFrame.removeCompletedDownloads();
					if (ServiceManager.sCfg.mUploadAutoRemoveCompleted)
					{
						mShareMgr.clearUpload();
					}
				}

				if (counter % 2 == 0)
				{
					if (mHostMgr.cleanupAndGetLiveHostCount() < ServiceManager.sCfg.mNetMinConn &&
						mNetworkMgr.getJoined())
					{
						mainFrame.autoConnectFromHostCatcher();
					}

					mDownloadMgr.retryDownloads();
					mHostMgr.checkConnections();
				}

				if (!mInitScan)
				{
					// Scan the file to share once at initialization time.
					mShareMgr.rescan();
					mInitScan = true;
				}

				if (!mInitSearch && counter > initSearchDuration && ServiceManager.sCfg.mAutoSearchCandidate)
				{
					// Perform auto search for download candidates at startup.
					mDownloadMgr.searchCandidates();
					mInitSearch = true;
				}

				counter++;
			}
			catch (Exception e)
			{
				// Catch exception and continue.  Don't kill the thread.
				e.printStackTrace();
			}
		}

		shutdownCompleted();
	}


	private synchronized void shutdownCompleted()
	{
		mRunning = false;
		ServiceManager.log("Background jobs runner stopped.");
		mShutdownCompleted = true;
		notifyAll();
	}

}


