
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex;


import java.net.*;
import java.io.*;
import java.util.*;

import phex.config.*;
import phex.host.*;
import phex.interfaces.*;


public class Listener implements Runnable
{
	private String				mSelfHostname = "127.0.0.1";
	private InetAddress			mListenerAddr = null;
	private int					mListeningPort = 0;
	private ServerSocket		mListeningSocket;
	private boolean				mRequestToDie = false;
	private boolean				mRunning = false;
	private boolean				mShutdownCompleted = false;
	private String				mMyServerAddress = "";
	private IHostChanged		mHostChangedListener = null;
	private InetAddress			mMyIP = null;




	public Listener()
	{
		try
		{
			mListenerAddr = InetAddress.getByName(mSelfHostname);
			if (ServiceManager.sCfg.mMyIP.length() > 0)
				mMyIP = InetAddress.getByName(ServiceManager.sCfg.mMyIP);
		}
		catch (Exception e)
		{
//			e.printStackTrace();
		}
	}


	public void setHostChangedListener(IHostChanged listener)
	{
		mHostChangedListener = listener;
	}


	public synchronized void startup()
			throws IOException
	{
		if (mRunning)
			return;

		// Init the ShareManager.
		ServiceManager.getShareManager();
		ServiceManager.log("Listener starting...");

		// Default to 0 for random port.
		mListeningPort = ServiceManager.sCfg.mListeningPort;

		// Create a listening socket at the port.
		mListeningSocket = new ServerSocket(mListeningPort, 50);

		mListeningPort = mListeningSocket.getLocalPort();

		mRequestToDie = false;
		mRunning = true;

		mSelfHostname = resolveSelfHostname();
		setSelfHostname(mSelfHostname);

		(new Thread(this)).start();
	}


	public synchronized void shutdown(boolean waitForCompleted)
	{
		// Already dead or been requested to die.
		if (mRequestToDie)
			return;

		ServiceManager.log("Listener shutting down...");

		// Set flag to die.
		mShutdownCompleted = false;
		mRequestToDie = true;

		// Dummy connection to the listener to wake it up to die.
		try
		{
			Socket	tmpSock = new Socket("127.0.0.1", mListeningPort);
			tmpSock.close();
		}
		catch(IOException e)
		{
			// Don't care.
		}

		if (waitForCompleted)
		{
			// Wait until the thread is dead.
			while (!mShutdownCompleted)
			{
				try
				{
					wait();
				}
				catch (InterruptedException e)
				{
					break;
				}
			}
		}
	}


	protected boolean getRequestToDie()
	{
		return mRequestToDie;
	}


	public boolean getRunning()
	{
		return mRunning;
	}


	protected String resolveSelfHostname()
	{
		InetAddress		addr = mListeningSocket.getInetAddress();
		String			hostAddress = null;

		// This could be slow.
//		hostAddress = addr.getHostName();
		if (hostAddress == null)
			hostAddress = addr.getHostAddress();
		if (hostAddress.equals("0.0.0.0"))
			return mSelfHostname;
		else
			return hostAddress;
	}


	public String getSelfHostname()
	{
		return mSelfHostname;
	}


	public void setSelfHostname(String hostname)
	{
		mSelfHostname = hostname;
		updateMyServerAddress();
	}


	private void updateMyServerAddress()
	{
		try
		{
			if (mMyIP == null)
			{
				mMyServerAddress = mSelfHostname + ":" + getListeningPort();
				mListenerAddr = InetAddress.getByName(mSelfHostname);
			}
			else
			{
				mMyServerAddress = mMyIP.getHostAddress() + ":" + getListeningPort();
				mListenerAddr = mMyIP;
			}

			if (mHostChangedListener != null)
				mHostChangedListener.updateMyIP();
		}
		catch (Exception e)
		{
//			e.printStackTrace();
		}
	}


	public void setMyIP(String ip)
	{
		try
		{
			if (ip.length() > 0)
				mMyIP = InetAddress.getByName(ip);
			else
				mMyIP = null;

			setSelfHostname(mSelfHostname);
		}
		catch (Exception e)
		{
			mMyIP = null;
		}
	}

			
	public InetAddress getSelfHostAddress()
	{
		return mListenerAddr;
	}


	public int getListeningPort()
	{
		if (mRunning)
			return mListeningPort;
		else
			return ServiceManager.sCfg.mListeningPort;
	}


	public String getMyServerAddress()
	{
		return mMyServerAddress;
	}



	// The listening thread.
	public void run()
	{
		ServiceManager.log("Listener started.");

		Socket	incoming;

		while (true)
		{
			try
			{
				// Waiting for incoming connection.
				incoming = mListeningSocket.accept();
				resolveSelfHostname();
			}
			catch(IOException e)
			{
				ServiceManager.log(e.toString());
				break;
			}

			// See if I has been asked to die.
			if (getRequestToDie())
			{
				break;
			}

			try
			{
				// Set this will defeat the Nagle Algorithm, making short bursts of
				// transmission faster, but will be worse for the overall network.
				// incoming.setTcpNoDelay(true);

				// Create a Host object for the incoming connection
				// and hand it off to a ReadWorker to handle.
				Host	host = new Host(ServiceManager.getHostManager());
				host.setType(Host.sTypeIncoming);
				host.setSock(incoming);
				host.setOs(incoming.getOutputStream());
				host.setIs(incoming.getInputStream());
				host.setHostAddr(incoming.getInetAddress().getHostAddress());
				host.setStatus(Host.sStatusAccepting, "");
				new ReadWorker(host, 0);
			}
			catch (Exception e)
			{
				ServiceManager.log(e.toString());
			}
		}

		try
		{
			mListeningSocket.close();
		}
		catch(IOException e)
		{
			ServiceManager.log(e.toString());
		}
		shutdownCompleted();
	}


	private synchronized void shutdownCompleted()
	{
		mRunning = false;
		ServiceManager.log("Listener stops.");
		mShutdownCompleted = true;
		notifyAll();
	}

}


