
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex.msg;


import java.util.*;

import phex.*;
import phex.config.*;
import phex.cqueue.*;
import phex.download.*;
import phex.host.*;
import phex.interfaces.*;
import phex.msg.*;


public class MsgManager
{
	private HostManager		mHostMgr = ServiceManager.getHostManager();
	private Hashtable		mMsgSeen;
	private Hashtable		mRoutingTable;
	private Hashtable		mPushRoutingTable;
	private CQueue			mMsgLRU;
	private CQueue			mRoutingLRU;
	private CQueue			mPushRoutingLRU;
	private MsgInit			mMyMsgInit;
	private Vector			mSearchList = new Vector();
	private Vector			mBackgroundSearchList = new Vector();
	private Vector			mPassiveResults = new Vector();
	private IMonitorSearch	mMonitorSearchListener;
	private FQueue			mNetSearchHistory = new FQueue(100);
	private String			mPassiveSearching = null;
	private Vector			mPassiveSearchingList = new Vector();


	public MsgManager()
	{
		mMsgSeen = new Hashtable();
		mRoutingTable = new Hashtable();
		mPushRoutingTable = new Hashtable();
		mMsgLRU = new CQueue(10000, false);		// hold upto 10000 msg
		mRoutingLRU = new CQueue(10000, false);	// hold upto 10000 routing guid
		mPushRoutingLRU = new CQueue(1000, false);	// hold upto 1000 routing guid

		mMyMsgInit = new MsgInit();
		mMyMsgInit.getHeader().setTTL(ServiceManager.sCfg.mNetTTL);

		// Add the InitMsg to the seen list.
		checkAndAddMsgSeen(mMyMsgInit);
	}


	public void setMonitorSearchListener(IMonitorSearch listener)
	{
		mMonitorSearchListener = listener;
	}



	// Test and set in one atomic operation because there are multiple
	// threads handling multiple connection at the same time.
	// If the test and set operations are done separately, there is
	// a chance that both threads would check and got the wrong result
	// before they can set.
	public synchronized boolean checkAndAddMsgSeen(IMsg msg)
	{
		GUID	msgID = msg.getHeader().getMsgID();

		if (mMsgSeen.get(msgID.getHashCode()) != null)
		{
			// Seen this msg.
			return true;
		}
		else
		{
			// Hasn't seen this msg.  Save it.
			mMsgSeen.put(msgID.getHashCode(), msg.getHeader());

			try
			{
				// Push into a FIFO LRU queue to keep track of the age of the object.
				mMsgLRU.addToHead(new CQueueElement(msgID));
				if (mMsgLRU.isExceedingMax())
				{
					// Too many guids.  Get rid of the oldest one.
					CQueueElement	oldestObj = (CQueueElement)mMsgLRU.removeFromTail();
					GUID			guid = (GUID)oldestObj.getObj();
					mMsgSeen.remove(guid.getHashCode());
				}
			}
			catch (Exception e)
			{
				e.printStackTrace();
			}

			return false;
		}
	}


	public synchronized void addToRoutingTable(GUID msgID, Host sender)
	{
		try
		{
			// Add to routing table.
			mRoutingTable.put(msgID.getHashCode(), sender);

			// Push into a FIFO LRU queue to keep track of the age of the object.
			mRoutingLRU.addToHead(new CQueueElement(msgID));
			if (mRoutingLRU.isExceedingMax())
			{
				// Too many guids.  Get rid of the oldest one.
				CQueueElement	oldestObj = (CQueueElement)mRoutingLRU.removeFromTail();
				GUID			guid = (GUID)oldestObj.getObj();
				mRoutingTable.remove(guid.getHashCode());
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

	}


	public synchronized void addToPushRoutingTable(GUID clientID, Host sender)
	{
		try
		{
			// Add the clientID in the search result of the sharing host
			// to routing table.
			mPushRoutingTable.put(clientID.getHashCode(), sender);

			// Push into a FIFO LRU queue to keep track of the age of the object.
			mPushRoutingLRU.addToHead(new CQueueElement(clientID));
			if (mPushRoutingLRU.isExceedingMax())
			{
				// Too many guids.  Get rid of the oldest one.
				CQueueElement	oldestObj = (CQueueElement)mPushRoutingLRU.removeFromTail();
				GUID			guid = (GUID)oldestObj.getObj();
				mPushRoutingTable.remove(guid.getHashCode());
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

	}


	public synchronized Host getRouting(GUID msgID)
	{
		return (Host)mRoutingTable.get(msgID.getHashCode());
	}


	public synchronized Host getPushRouting(GUID clientID)
	{
		return (Host)mPushRoutingTable.get(clientID.getHashCode());
	}


	public MsgInit getMyMsgInit()
	{
		return mMyMsgInit;
	}


	public void resetMyMsgInit()
	{
		mMyMsgInit = new MsgInit();
	}


	public int getSearchCount()
	{
		return mSearchList.size();
	}


	public Search getSearch(int index)
	{
		if (index > getSearchCount() - 1)
			return null;
		return (Search)mSearchList.elementAt(index);
	}


	public synchronized void sendSearch(int minSpeed, String queryStr, ISearch searchListener)
	{
		Search	search = new Search(minSpeed, queryStr, 0, searchListener);
		mSearchList.insertElementAt(search, 0);
		if (mSearchList.size() > ServiceManager.sCfg.mSearchMaxConcurrent)
			mSearchList.removeElementAt(mSearchList.size() - 1);
		search.startSearching();
	}


	public synchronized Search sendBackgroundSearch(String queryStr, ISearch searchListener)
	{
		Search	search = new Search(0, queryStr, 0, searchListener);
		mBackgroundSearchList.addElement(search);
		if (mBackgroundSearchList.size() > ServiceManager.sCfg.mMaxDownload + 2)
			mBackgroundSearchList.removeElementAt(0);
		search.startSearching();
		return search;
	}


	public synchronized void stopBackgroundSearch(Search search)
	{
		search.stopSearching();
		mBackgroundSearchList.removeElement(search);
	}


	public synchronized void processQueryResponse(Host remoteHost, MsgQueryResponse msg)
	{
		for (int i = 0; i < mSearchList.size(); i++)
		{
			((Search)mSearchList.elementAt(i)).processResponse(remoteHost, msg);
		}
		for (int i = 0; i < mBackgroundSearchList.size(); i++)
		{
			((Search)mBackgroundSearchList.elementAt(i)).processResponse(remoteHost, msg);
		}

		if (!mHostMgr.isHostFiltered(msg.getRemoteHost().getHostAddress()))
		{
			passiveMonitorResult(msg);
		}
		
	}


	public void stopAllSearches()
	{
		for (int i = 0; i < mSearchList.size(); i++)
		{
			((Search)mSearchList.elementAt(i)).stopSearching();
		}
	}


	public void removeAllSearches()
	{
		stopAllSearches();
		mSearchList.removeAllElements();
	}


	public FQueue getNetSearchHistory()
	{
		return mNetSearchHistory;
	}


	public void addNetSearch(MsgQuery msg)
	{
		if (ServiceManager.sCfg.mSearchEnableMonitor)
		{
			return;
		}

		mNetSearchHistory.add(msg);
		mMonitorSearchListener.searchArrived();
	}


	public String getPassiveSearching()
	{
		return mPassiveSearching;
	}


	public void setPassiveSearching(String str)
	{
		mPassiveSearching = str;
		mPassiveSearchingList.removeAllElements();

		if (str != null)
		{
			StringTokenizer	tokens = new StringTokenizer(str);
			while (tokens.hasMoreElements())
			{
				mPassiveSearchingList.addElement(tokens.nextElement());
			}
		}
	}


	public Vector getPassiveResults()
	{
		return mPassiveResults;
	}


	public void passiveMonitorResult(MsgQueryResponse msg)
	{
		if (mPassiveSearching == null)
			return;

		// resolve: have its own property name.
		if (mPassiveResults.size() >= ServiceManager.sCfg.mSearchMaxSearch)
		{
			// resolve: reset status
			return;
		}

		String	hostStr = msg.getRemoteHostStr();
		int		speed = msg.getRemoteHostSpeed();

		for (int i = 0; i < msg.getNumRecords(); i++)
		{
			MsgResRecord	rec = msg.getMsgRecord(i);

			// Capture everything.
			if (mPassiveSearching.length() == 0)
			{
				RemoteFile	rfile = new RemoteFile(
					msg.getRemoteClientID(),
					rec.getFileIndex(),
					rec.getFilename(),
					rec.getFileSize(),
					hostStr,
					speed);

				if (!passiveResultExists(rfile))
				{
					mPassiveResults.addElement(rfile);
					mMonitorSearchListener.passiveResultArrived();
					continue;
				}
			}

			// Capture using only the filter match keyword.
			for (int j = 0; j < mPassiveSearchingList.size(); j++)
			{
				String		filter = (String)mPassiveSearchingList.elementAt(j);
				if (rec.getFilename().indexOf(filter) != -1)
				{
					RemoteFile	rfile = new RemoteFile(
						msg.getRemoteClientID(),
						rec.getFileIndex(),
						rec.getFilename(),
						rec.getFileSize(),
						hostStr,
						speed);

					if (!passiveResultExists(rfile))
					{
						mPassiveResults.addElement(rfile);
						mMonitorSearchListener.passiveResultArrived();
					}
					break;
				}
			}
		}

	}


	// resolve: this is expensive.
	private boolean passiveResultExists(RemoteFile rfile)
	{
		for (int i = 0; i < mPassiveResults.size(); i++)
		{
			RemoteFile	pfile = (RemoteFile)mPassiveResults.elementAt(i);
			if (pfile.equals(rfile))
				return true;
		}
		return false;
	}


	public void resetPassiveResults()
	{
		mPassiveResults.removeAllElements();
		mMonitorSearchListener.passiveResultArrived();
	}



}

