
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex;


import java.io.*;
import java.net.*;
import java.util.*;

import phex.config.*;
import phex.connection.*;
import phex.download.*;
import phex.host.*;
import phex.interfaces.*;
import phex.msg.*;
import phex.utils.*;


public class ReadWorker implements Runnable
{
	// GNUTELLA CONNECT/0.4\n\n
	private static final String	sRequestSignature = "GNUTELLA CONNECT/0.4";
	private static final String	sReplySignature = "GNUTELLA OK";
	private static final String	sPasswordChallege = "GNUTELLA AUTHENTICATION CHALLENGE";
	private static final String	sPasswordReply = "GNUTELLA AUTHENTICATION RESPONSE";
	private static final String	sAuthFailed = "GNUTELLA AUTHENTICATION FAILED";
	
	private Host			mRemoteHost;
	private byte[]			mHeaderBuf;		// pre-allocated buffer for repeated uses.
	private FlexBuf			mBuf1;			// general buf for read/send
	private FlexBuf			mBuf2;			// general buf for read/send
	private FlexBuf			mBodyBuf;		// Buf for reading msg body
	private int				mSleep = 1;
	private int				mSleepBeforeStart = 0;
	private ServiceManager	mManager = ServiceManager.getManager();
	private ConnectionManager	mConnMgr = ServiceManager.getConnectionManager();
	private HostManager		mHostMgr = mManager.getHostManager();
	private SendManager		mSendMgr = mManager.getSendManager();
	private ShareManager	mShareMgr = mManager.getShareManager();
	private DownloadManager	mDownloadMgr = mManager.getDownloadManager();
	private MsgManager		mMsgMgr = mManager.getMsgManager();
//	private IrcManager		mIrcMgr = mManager.getIrcManager();


	private ReadWorker()
	{
		// disable
	}


	public ReadWorker(Host remoteHost, int sleepBeforeStart)
	{
		mRemoteHost = remoteHost;
		mHeaderBuf = new byte[MsgHeader.sDataLength];
		mBuf1 = new FlexBuf();
		mBuf2 = new FlexBuf();
		mBodyBuf = new FlexBuf();
		mSleepBeforeStart = sleepBeforeStart;

		(new Thread(this)).start();
	}


	public void run()
	{
		if (!mRemoteHost.acquireByWorker())
		{
			// Already has a worker working on it.  Just exit.
			return;
		}

		if (mSleepBeforeStart > 0)
		{
			try
			{
				Thread.sleep(mSleepBeforeStart);
			}
			catch (Exception e)
			{
			}
		}

		try
		{
			if (mRemoteHost.getType() == Host.sTypeOutgoing)
			{
				// Connect to remote host.
				mRemoteHost.setStatus(Host.sStatusConnecting, "");

				try
				{
					connectToRemoteHost();
					if (mRemoteHost.getStatus() == Host.sStatusTimeout)
					{
						// Connecting has been taken too long.
						throw new Exception("Timed out.");
					}
				}
				catch (Exception e3)
				{
					mHostMgr.setHostCaughtConnectionFailed(mRemoteHost.getHostAddr(), true);
					throw e3;
				}
				
				// Set my local address.
				String	localaddr = mRemoteHost.getSock().getLocalAddress().getHostAddress();
				mManager.getListener().setSelfHostname(localaddr);

				// Negotiate handshake
				mRemoteHost.setStatus(Host.sStatusConnecting, "Negotiate handshake.");
				negotiateHandShakeAsClient();

				// Connection to remote gnutella host is completed at this point.
				mRemoteHost.setStatus(Host.sStatusConnected, "");
				mHostMgr.addConnectedHost(mRemoteHost);

				// queue first Init msg to send.
				mSendMgr.queueMsgToSend(mRemoteHost, mMsgMgr.getMyMsgInit(), false);

				processIncomingData();
			}
			else if (mRemoteHost.getType() == Host.sTypeIncoming)
			{
				ConnectionRequest	req = negotiateHandShakeAsServer();

				// See if the incoming connection is from a gnutella client.
				if (req.getMethod().equals(ConnectionRequest.sGnutellaConnect))
				{
					// Incoming connection is from a gnutella client.
					if (mHostMgr.getConnectedHostCount() >= ServiceManager.sCfg.mNetMaxConnection)
					{
						// Exceed total connection limit (both incoming and outgoing).
						// Refuse additional incoming connection.
						return;
					}

					checkPassword();

					byte[]	buf = mBuf1.getBuf(1024);
					int		len;

					// Send reply signature back to remote host
					len = IOUtil.serializeString(sReplySignature + "\n\n", buf, 0);
					mRemoteHost.getOs().write(buf, 0, len);

					mHostMgr.addIncomingHost(mRemoteHost);
					mHostMgr.addConnectedHost(mRemoteHost);

					processIncomingData();
				}
				else if (req.getMethod().equals(ConnectionRequest.sGet))
				{
					// Incoming connection is a HTTP GET to upload file.
					mShareMgr.httpRequestHandler(req, mRemoteHost);
				}
				else if (req.getMethod().equals(ConnectionRequest.sGiv))
				{
					// Incoming connection is a GIV, a response to my push request.
					// The remote host is pushing (sending) the file to me now.
					// Transfer mRemoteHost and its connection to the download thread.
					if (mDownloadMgr.notifyReadPushedFile(req, mRemoteHost))
					{
						// mRemoteHost has been transferred.  Relinquish control of it.
						mRemoteHost.releaseFromWorker();
						mRemoteHost = null;
					}
				}
				else
				{
					throw new Exception("Unknown incoming request type.");
				}
			}
			else
			{
				throw new Exception("Invalid host type");
			}
		}
		catch (Exception e)
		{
			mRemoteHost.setStatus(Host.sStatusError, e.getMessage());
//			e.printStackTrace();
		}
		finally
		{
			// Close the connection.  Need to close even for error.
			// Half-closed connection is bad for TCP.
			if (mRemoteHost != null)
			{
				mRemoteHost.closeConnection();
				mRemoteHost.releaseFromWorker();
			}
		}
	}


	private void connectToRemoteHost()
			throws Exception
	{
		try
		{
			mRemoteHost.markConnectionStartTime();
			Socket	sock = mConnMgr.connect(mRemoteHost.getHostAddr());

			// Set this will defeat the Nagle Algorithm.  Make short bursts of
			// transmission faster, but will be worse for the overall LAN.
//			sock.setTcpNoDelay(true);

			// I am connected to the remote host at this point.
			mRemoteHost.setSock(sock);
			mRemoteHost.setOs(sock.getOutputStream());
			mRemoteHost.setIs(sock.getInputStream());
		}
		catch (UnknownHostException e)
		{
			throw new IOException("Unknown host.");
		}
	}


	private void negotiateHandShakeAsClient()
			throws Exception
	{
		// Send the first handshake greeting to the remote host.
		boolean	sentPassword = false;
		String	greeting = ServiceManager.getNetworkManager().getFullGreeting() + "\n\n";
		byte[]	buf = mBuf1.getBuf(1024);
		int		len = IOUtil.serializeString(greeting, buf, 0);
		mRemoteHost.getOs().write(buf, 0, len);

		try
		{
			while (true)
			{
				// Read reply from server.
				StringBuffer	strBuf = new StringBuffer();
				len = IOUtil.readToCRLF(mRemoteHost.getIs(), buf, 1024, 0);
				IOUtil.deserializeString(buf, 0, len, strBuf);
				String			reply = strBuf.toString();
				String			password = (String)ServiceManager.sCfg.mNetworkPasswords.get(ServiceManager.sCfg.mCurrentNetwork);

				if (reply.equals(sReplySignature))
				{
					if (!sentPassword && password != null && password.length() > 0)
					{
						// I have password set on this network name but the
						// remote host doesn't have for any password.
						// Don't want to join a unsecured network.
						throw new Exception("Remote host doesn't have password set.");
					}

					// Normal Gnutella reply.
					return;
				}
				else if (reply.startsWith(sPasswordChallege))
				{
					// Remote host challenges for password.
					String		keyStr = reply.substring(sPasswordChallege.length() + 1);
					byte[]		key = HexDec.convertHexStringToBytes(keyStr);

					if (password == null || password.length() == 0)
					{
						throw new Exception("Password requried.");
					}

					byte[]		authDigest = Digest.computePasswordDigest(password, key);
					String		authDigestStr = HexDec.convertBytesToHexString(authDigest);
			
					// Send password reply.
					len = IOUtil.serializeString(sPasswordReply + " " + authDigestStr + "\n\n", buf, 0);
					mRemoteHost.getOs().write(buf, 0, len);
					sentPassword = true;

					// Loopback to read normal reply.
					continue;
				}
				else if (reply.startsWith(sAuthFailed))
				{
					throw new Exception("Password not accepted by remote host.");
				}

				// I don't recognize any other reply.
				throw new Exception(reply);
			}
		}
		catch (Exception e)
		{
			if (e instanceof IOException)
				throw new IOException("Disconnected from remote host during initial handshake");
			throw e;
		}
	}


	private ConnectionRequest negotiateHandShakeAsServer()
			throws Exception
	{
		if (!ServiceManager.getNetworkManager().getJoined())
		{
			throw new Exception("Network not joined.");
		}
		if (mHostMgr.isHostIgnored(mRemoteHost.getHostAddr()))
		{
			throw new Exception("Host is ignored.");
		}

		// GNUTELLA CONNECT/0.4\n\n
		// GET /get/1/foo.txt HTTP/1.1\r\n
		// GIV <file-ref-num>:<ClientID GUID in hexdec>/<filename>\n\n
		InputStream		is = mRemoteHost.getIs();
		if (is == null)
		{
			// IS from getIs() has been cleared when connection closed.
			throw new IOException("Disconnected from remote host during initial handshake");
		}

		byte[]			buf = mBuf1.getBuf(1024);
		int				lenRead = IOUtil.readToCRLF(is, buf, 1024, 0);
		StringBuffer	strBuf = new StringBuffer();
		IOUtil.deserializeString(buf, 0, lenRead, strBuf);

		return (new ConnectionRequest(strBuf.toString()));

	}


	private void checkPassword()
			throws Exception
	{
		String	currPassword = (String)ServiceManager.sCfg.mNetworkPasswords.get(ServiceManager.sCfg.mCurrentNetwork);
		
		if (currPassword != null && currPassword.length() > 0)
		{
			StringBuffer	strBuf = new StringBuffer();
			byte[]			buf = mBuf1.getBuf(1024);
			byte[]			key = Digest.generateChallengeKey();
			String			keyStr = HexDec.convertBytesToHexString(key);
			byte[]			myDigest = Digest.computePasswordDigest(currPassword, key);
			String			myDigestStr = HexDec.convertBytesToHexString(myDigest);
			
			int				len = IOUtil.serializeString(sPasswordChallege + " " + keyStr + "\n\n", buf, 0);
			mRemoteHost.getOs().write(buf, 0, len);

			len = IOUtil.readToCRLF(mRemoteHost.getIs(), buf, 1024, 0);
			IOUtil.deserializeString(buf, 0, len, strBuf);
			String			reply = strBuf.toString();

			if (reply.startsWith(sPasswordReply))
			{
				String		clientDigestStr = reply.substring(sPasswordReply.length() + 1);
				clientDigestStr.toUpperCase();

				if (!myDigestStr.equals(clientDigestStr))
				{
					len = IOUtil.serializeString(sAuthFailed + "\n\n", buf, 0);
					mRemoteHost.getOs().write(buf, 0, len);

					// Wait a bit before closing the connection.
					// Somehow the remote gnutella won't read the last
					// buffer if closing the connection too soon.
					Thread.sleep(2000);

					throw new Exception("Client password not matching.");
				}
			}
		}
	}


	private void processIncomingData()
			throws Exception
	{
		mRemoteHost.setStatus(Host.sStatusConnected, "");

		try
		{
			while (true)
			{
				MsgHeader	header = readHeader();
				if (header.getDataLen() > 65536)
				{
					// Packet looks suspiciously too big.  Fuck it.  Disconnect.
					mRemoteHost.log("Packet too big.  " + header);
					throw new Exception("Packet too big.  Disconnecting the remote host.");
				}

				byte[]		body = readBody(header);

				mRemoteHost.incReceivedCount();
				mHostMgr.incStatMsgCount(1);
				mHostMgr.incStatTakenCount(header.getHopsTaken());

				switch (header.getFunction())
				{
					case MsgHeader.sInit:
						handleInit(header, body);
						break;

					case MsgHeader.sInitResponse:
						handleInitResponse(header, body);
						break;

					case MsgHeader.sPushRequest:
						handlePushRequest(header, body);
						break;

					case MsgHeader.sQuery:
						handleQuery(header, body);
						break;

					case MsgHeader.sQueryResponse:
						handleQueryResponse(header, body);
						break;

					default:
						handleUnknown(header, body);
						break;
				}
			}
		}
		catch (Exception e)
		{
//			e.printStackTrace();
			if (mRemoteHost.getSock() != null)
			{
				mRemoteHost.setStatus(Host.sStatusError, e.getMessage());
			}
		}
	}


	private MsgHeader readHeader()
			throws Exception
	{
		int			lenRead = 0;
		int			len;
		InputStream	is = mRemoteHost.getIs();

		if (is == null)
		{
			// IS from getIs() has been cleared when connection closed.
//			mRemoteHost.log("Disconnected from remote host during reading header");
			throw new Exception("Connection closed by remote host");
		}

		while (lenRead < MsgHeader.sDataLength)
		{
			len = is.read(mHeaderBuf, lenRead, MsgHeader.sDataLength - lenRead);
			if (len == 0 || len == -1)
			{
//				mRemoteHost.log("Disconnected from remote host during reading header");
				throw new Exception("Connection closed by remote host");
			}
			lenRead += len;
		}
		mHostMgr.incBytesCount(lenRead);

		// resolve: reuse msg header from free list.
		MsgHeader	header = new MsgHeader(new GUID(true));

		header.deserialize(mHeaderBuf, 0);
		header.setArrivalTime(System.currentTimeMillis());
		header.setFromHost(mRemoteHost);

		mSleep = mHostMgr.throttleControl(mSleep, 4000, mHostMgr.getRate(),
										  ServiceManager.sCfg.mNetMaxRate);

		return header;
	}


	private byte[] readBody(MsgHeader header)
			throws Exception
	{
		int			lenToRead = header.getDataLen();
		int			lenRead = 0;
		int			len;
		int			size;
		byte[]		body = mBodyBuf.getBuf(lenToRead);
		InputStream	is = mRemoteHost.getIs();

		if (is == null)
		{
			// IS from getIs() has been cleared when connection closed.
//			mRemoteHost.log("Disconnected from remote host during reading msg body");
			throw new Exception("Connection closed by remote host");
		}

		while (lenRead < lenToRead)
		{
			size = lenToRead - lenRead;
			if (size > 1024)
				size = 1024;
			len = is.read(body, lenRead, size);
			if (len == 0 || len == -1)
			{
//				mRemoteHost.log("Disconnected from remote host during reading message body.");
				throw new Exception("Connection closed by remote host");
			}
			lenRead += len;

			mHostMgr.incBytesCount(len);
			mSleep = mHostMgr.throttleControl(mSleep, 4000, mHostMgr.getRate(),
											  ServiceManager.sCfg.mNetMaxRate);
		}

		return body;
	}


	private void handleInit(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgInit			msg = new MsgInit(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got msg " + msg);

		// See if I have seen this Init before.  Drop msg if duplicate.
		if (mMsgMgr.checkAndAddMsgSeen(msg))
		{
			mRemoteHost.incDropCount();
			mHostMgr.incStatDropCount(1);
//			mRemoteHost.log("Seen " + msg);
			return;
		}

		// Add the Init msg to the routing table so that I know where
		// to route the InitResponse back.
		mMsgMgr.addToRoutingTable(msg.getHeader().getMsgID(), mRemoteHost);

		// Forward the Init msg to other connected hosts except the one sent it.
		mHostMgr.forwardMsg(msg, mRemoteHost);

		// Get my host:port for InitResponse.
		Listener	listener = mManager.getListener();

		// Construct InitResponse msg.  Copy the original Init's GUID.
		MsgHeader	newHeader = new MsgHeader(header.getMsgID());
		MsgInitResponse	response = new MsgInitResponse(newHeader);
		newHeader.setTTL(header.getHopsTaken() + 1);	// Will take as many hops to get back.
		response.setPort((short)listener.getListeningPort());
		response.setIP(listener.getSelfHostAddress());
		response.setFileCount(mShareMgr.getFileCount());
		response.setTotalSize(mShareMgr.getTotalFileSize());

		mSendMgr.queueMsgToSend(mRemoteHost, response, false);
	}


	private void handleInitResponse(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgInitResponse	msg = new MsgInitResponse(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got msg " + msg);

		if (!mHostMgr.isHostInvalid(msg.getIP().getHostAddress()))
		{
			mHostMgr.addAndSaveHostCaught(msg.getIP().getHostAddress() + ":" + msg.getPort());
		}

		// Handle init response to my origianl init.
		MsgInit			msgInit = mMsgMgr.getMyMsgInit();
		if (msg.getHeader().getMsgID().equals(msgInit.getHeader().getMsgID()))
		{
			// This InitResponse is for me.
//			mRemoteHost.log("Got response to my init.  " + msg);
			mHostMgr.incStatHosts(1);
			mHostMgr.incStatFiles(msg.getFileCount());
			mHostMgr.incStatSize(msg.getTotalSize());
			return;
		}

		// See if the response is to my latency ping to neighbor.
		if (mRemoteHost.checkPingResponse(msg))
		{
			// Yes.  Processed.
			return;
		}

		// InitResponse is a response to an Init msg.
		// Did I forward that Init msgID before?
		Host			returnHost = mMsgMgr.getRouting(msg.getHeader().getMsgID());
		if (returnHost == null)
		{
//			mRemoteHost.log("Don't route the InitResponse since I didn't forward the Init msg.  " + msg);
			return;
		}

		// Ok, I did forward the Init msg on behalf of returnHost.
		// The InitResponse is for returnHost.  Better route it back.
		if (mHostMgr.decTTL(msg))
		{
//			mRemoteHost.log("TTL expired " + msg);
			return;
		}
		mSendMgr.queueMsgToSend(returnHost, msg, false);
	}


	private void handleQuery(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgQuery		msg = new MsgQuery(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got msg " + msg);

		// See if I have seen this Query before.  Drop msg if duplicate.
		if (mMsgMgr.checkAndAddMsgSeen(msg))
		{
			mRemoteHost.incDropCount();
			mHostMgr.incStatDropCount(1);
//			mRemoteHost.log("Seen " + msg);
			return;
		}

		// Add the Query msg to the routing table so that I know where
		// to route the QueryResponse back.
		mMsgMgr.addToRoutingTable(msg.getHeader().getMsgID(), mRemoteHost);

		// Add to the net search history.
		mMsgMgr.addNetSearch(msg);

		// Forward the Query msg to other connected hosts except the one sent it.
		mHostMgr.forwardMsg(msg, mRemoteHost);

		// Perform search on my list.
		if (msg.getMinSpeed() > ServiceManager.sCfg.mNetMySpeed)
		{
			return;
		}

		// Search the sharefile database and get groups of sharefiles.
		Hashtable		groups = mShareMgr.searchFile(msg.getSearchString());

		if (groups.size() == 0)
		{
//			mRemoteHost.log("Not found");
			return;
		}

		Listener		listener = mManager.getListener();

		// For each group, build a separate MsgQueryResponse packet
		// since the host:port for each group is different.
		Enumeration		groupEnum = groups.elements();
		while (groupEnum.hasMoreElements())
		{
			// Get list of sharefile in each group.
			Vector				list = (Vector)groupEnum.nextElement();

			// Construct QueryResponse msg.  Copy the original Init's GUID.
			MsgHeader			newHeader = new MsgHeader(header.getMsgID());
			MsgQueryResponse	response = new MsgQueryResponse(newHeader);
			int					recCount = list.size();

			if (recCount == 0)
				continue;
			if (recCount > 255)
				recCount = 255;

			newHeader.setTTL(header.getHopsTaken() + 1);	// Will take as many hops to get back.
			response.setNumRecords(recCount);
			response.setRemoteHostSpeed(ServiceManager.sCfg.mNetMySpeed);
			response.setRemoteClientID(mManager.getClientID());

			ShareFile			sfile = null;
			for (int i = 0; i < recCount; i++)
			{
				sfile = (ShareFile)list.elementAt(i);
				if (sfile.getRemoteShareFileHost() == null)
				{
					// Local share file
					response.addMsgRecord(new MsgResRecord(sfile.getFileIndex().intValue(), (int)sfile.getLength(), sfile.getEffectiveName()));
				}
				else
				{
					// Remote share file
					response.addMsgRecord(new MsgResRecord(sfile.getRemoteShareFileNumber(), (int)sfile.getLength(), sfile.getEffectiveName()));
				}
			}

			if (sfile.getRemoteShareFileHost() == null)
			{
				// Local share file
				response.setRemoteListeningPort((short)listener.getListeningPort());
				response.setRemoteHost(listener.getSelfHostAddress());
			}
			else
			{
				// Remote share file
				response.setRemoteListeningPort((short)sfile.getRemoteShareFilePort());
				response.setRemoteHost(sfile.getRemoteShareFileHost());
			}

			// Send each batch in sequence.
			mSendMgr.queueMsgToSend(mRemoteHost, response, false);
		}
	}


	private void handleQueryResponse(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgQueryResponse	msg = new MsgQueryResponse(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got msg " + msg);

		mMsgMgr.addToPushRoutingTable(msg.getRemoteClientID(), mRemoteHost);
		mMsgMgr.processQueryResponse(mRemoteHost, msg);

		// QueryResponse is a response to a Query msg.
		// Did I forward that Query msgID before?
		Host			returnHost = mMsgMgr.getRouting(msg.getHeader().getMsgID());
		if (returnHost == null)
		{
//			mRemoteHost.log("Don't route the QueryResponse since I didn't forward the Query msg.  " + msg);
			return;
		}

		// Ok, I did forward the Query msg on behalf of returnHost.
		// The QueryResponse is for returnHost.  Better route it back.
		if (mHostMgr.decTTL(msg))
		{
//			mRemoteHost.log("TTL expired " + msg);
			return;
		}
		mSendMgr.queueMsgToSend(returnHost, msg, false);
	}


	private void handlePushRequest(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgPushRequest		msg = new MsgPushRequest(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got msg " + msg);

		if (mManager.getClientID().equals(msg.getClientID()))
		{
			new PushWorker(msg);
			return;
		}

		Host	returnHost = mMsgMgr.getPushRouting(msg.getClientID());
		if (returnHost == null)
		{
//			mRemoteHost.log("Don't route the PushRequest since I didn't forward the QueryResponse msg.  " + msg);
			return;
		}

		// Ok, I did forward the QueryResponse msg on behalf of returnHost.
		// The PushRequest is for the returnHost.  Better route it back.
		if (mHostMgr.decTTL(msg))
		{
//			mRemoteHost.log("TTL expired " + msg);
			return;
		}
		mSendMgr.queueMsgToSend(returnHost, msg, false);
	}


	private void handleUnknown(MsgHeader header, byte[] body)
			throws Exception
	{
		MsgUnknown		msg = new MsgUnknown(header);

		msg.deserialize(body, 0);
//		mRemoteHost.log("Got unknown msg " + msg);


		// See if I have seen this msg before.  Drop msg if duplicate.
		if (mMsgMgr.checkAndAddMsgSeen(msg))
		{
			mRemoteHost.incDropCount();
			mHostMgr.incStatDropCount(1);
//			mRemoteHost.log("Seen " + msg);
			return;
		}

		// Forward the msg to other connected hosts except the one sent it.
		mHostMgr.forwardMsg(msg, mRemoteHost);
	}

}


