
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex;


import java.net.*;
import java.io.*;
import java.util.*;

import phex.config.*;
import phex.connection.*;
import phex.download.*;
import phex.host.*;
import phex.msg.*;
import phex.utils.*;


public class ServiceManager
{
	static String					sConfigPathname = System.getProperty("user.home") +
												File.separator + "phex.cfg";
	private static String			sLogFilename = System.getProperty("user.home") +
												File.separator + "phex.log";
	private static String			sHostsPathname = System.getProperty("user.home") +
												File.separator + "phex.hosts";
	private static String			sDownloadSavePathname = System.getProperty("user.home") +
												File.separator + "phex.download";
	private static ServiceManager	sTheManager = null;
	private static Object			sInitLocker = new Object();

	public static Cfg				sCfg;

	private String					mLogfilename;
	private String					mHostsFilename;
	private String					mDownloadSaveFilename;
	private PrintWriter				mLogFile;
	private ConnectionManager		mConnectionManager;
	private NetworkManager			mNetworkManager;
	private SendManager				mSendManager;
	private HostManager				mHostManager;
	private MsgManager				mMsgManager;
	private ShareManager			mShareManager;
	private DownloadManager			mDownloadManager;
	private IndexerManager			mIndexerManager;
	private Listener				mListener;
//	private Identd					mIdentd;
	private JobsRunner				mJobsRunner;
	private MainFrame				mMainFrame;
//	private IrcConsoleFrame			mIrcConsoleFrame;


	public static void setConfigFilename(String filename)
	{
		sConfigPathname = filename;
	}


	public static ServiceManager getManager()
	{
		
		// Simple and fast check to avoid taking a mutex lock.
		if (sTheManager == null)
		{
			// Mutex to allow only one thread to initialize the manager.
			synchronized (sInitLocker)
			{
				// Check again in case someone else is initializing the manager.
				if (sTheManager == null)
				{
					sTheManager = new ServiceManager();
					sTheManager.init();
				}
			}
		}

		return sTheManager;
	}


	private ServiceManager()
	{
		sCfg = new Cfg(sConfigPathname);
		sCfg.load();

		// Set up log file.
		mLogfilename = sCfg.mLogFile;
		if (mLogfilename == null || mLogfilename.length() == 0)
			mLogfilename = sLogFilename;
		mLogfilename = envSubstitution(mLogfilename);

		try
		{
			mLogFile = new PrintWriter(new BufferedWriter(new FileWriter(mLogfilename)));
		}
		catch (IOException e)
		{
			mLogFile = new PrintWriter(new OutputStreamWriter(System.out));
			logMsg("Failed to create default log file '" + mLogfilename + ".'  " + e.toString());
			logMsg("Use standard output.");
		}

		// Set up hosts file.
		mHostsFilename = sCfg.mHostFile;
		if (mHostsFilename == null || mHostsFilename.length() == 0)
			mHostsFilename = sHostsPathname;
		mHostsFilename = envSubstitution(mHostsFilename);

		// Set up the filename to save download data.
		mDownloadSaveFilename = sCfg.mDownloadSaveFile;
		if (mDownloadSaveFilename == null || mDownloadSaveFilename.length() == 0)
			mDownloadSaveFilename = sDownloadSavePathname;
		mDownloadSaveFilename = envSubstitution(mDownloadSaveFilename);
	}


	private void init()
	{
		log("\n*** ServiceManager initializing ***");

		mIndexerManager = new IndexerManager(this);
		mConnectionManager = new ConnectionManager();
		mNetworkManager = new NetworkManager();
		mHostManager = new HostManager();
		mSendManager = new SendManager();
		mMsgManager = new MsgManager();
		mShareManager = new ShareManager();
		mDownloadManager = new DownloadManager();
		mListener = new Listener();
//		mIdentd = new Identd();
		mJobsRunner = new JobsRunner();

		log("*** ServiceManager is ready ***");
	}


	public synchronized void shutdown()
	{
		mLogFile.flush();

		log("manager shutting down...");

		mJobsRunner.shutdown(true);

		mLogFile.close();
	}


	public void startJobsRunner()
	{
		mJobsRunner.startup();
	}


	public static ConnectionManager getConnectionManager()
	{
		return ServiceManager.getManager().mConnectionManager;
	}


	public static NetworkManager getNetworkManager()
	{
		return ServiceManager.getManager().mNetworkManager;
	}


	public static SendManager getSendManager()
	{
		return ServiceManager.getManager().mSendManager;
	}


	public static HostManager getHostManager()
	{
		return ServiceManager.getManager().mHostManager;
	}


	public static MsgManager getMsgManager()
	{
		return ServiceManager.getManager().mMsgManager;
	}


	public static ShareManager getShareManager()
	{
		return ServiceManager.getManager().mShareManager;
	}


	public static DownloadManager getDownloadManager()
	{
		return ServiceManager.getManager().mDownloadManager;
	}


	public static IndexerManager getIndexerManager()
	{
		return ServiceManager.getManager().mIndexerManager;
	}


	public static Listener getListener()
	{
		return ServiceManager.getManager().mListener;
	}

/*
	public static Identd getIdentd()
	{
		return ServiceManager.getManager().mIdentd;
	}
*/
	public GUID getClientID()
	{
		return sCfg.mProgramClientID;
	}


	public static String getHostsFilename()
	{
		if (sCfg.mCurrentNetwork.equals(sCfg.sGeneralNetwork))
		{
			return getManager().mHostsFilename;
		}
		else
		{
			String	filename = "phex_" + sCfg.mCurrentNetwork + ".hosts";
			filename = StrUtil.convertToLocalSystemFilename(filename);
			return System.getProperty("user.home") + File.separator + filename;
		}
	}


	public static String getDownloadSaveFilename()
	{
		return getManager().mDownloadSaveFilename;
	}


	public static void log(String msg)
	{
		ServiceManager.getManager().logMsg(msg);
	}


	public static void log(Exception e)
	{
		ServiceManager.getManager().logMsg(e);
	}


	public void logMsg(String msg)
	{
		mLogFile.println(msg);
		mLogFile.flush();
	}


	public void logMsg(Exception e)
	{
		mLogFile.println(e.toString());
		e.printStackTrace(mLogFile);
		mLogFile.flush();
	}


	public static String envSubstitution(String str)
	{
		String	str2 = str.toLowerCase();
		String	strEnv1 = "$user.home";
		String	strEnv2 = "$(user.home)";

		if (str2.startsWith(strEnv1))
		{
			str = System.getProperty("user.home") +
				  str.substring(strEnv1.length());
		}
		else if (str2.startsWith(strEnv2))
		{
			str = System.getProperty("user.home") +
				  str.substring(strEnv2.length());
		}

		return (str);
	}


	void setMainFrame(MainFrame frame)
	{
		mMainFrame = frame;
	}


	public MainFrame getMainFrame()
	{
		return mMainFrame;
	}

/*
	void setIrcConsoleFrame(IrcConsoleFrame frame)
	{
		mIrcConsoleFrame = frame;
	}


	synchronized IrcConsoleFrame getIrcConsoleFrame()
	{
		if (mIrcConsoleFrame == null)
		{
			mIrcConsoleFrame = new IrcConsoleFrame(null);
		}
		return mIrcConsoleFrame;
	}

*/
	public static void exec(String cmd)
			throws Exception
	{
		Runtime		rt = Runtime.getRuntime();

		rt.exec(cmd);
	}


	public static void printStack(String str)
	{
		try
		{
			throw new Exception(str);
		}
		catch (Exception e)
		{
			log(e);
		}
	}

}


