
/*
 *  PHEX - The pure-java Gnutella-servant.
 *  Copyright (C) 2000 William W. Wong
 *  williamw@jps.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package phex;


import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;

import phex.config.*;
import phex.connection.*;
import phex.host.*;
import phex.interfaces.*;
import phex.utils.*;


public class ShareManager
{
	private final static String		sAliasFile = "filealias";
	private final static String		sAliasFilename = sAliasFile + ".txt";

	private Vector			mDirsToShare = new Vector();
	private Vector			mFilesToShare = new Vector();
	private Hashtable		mNumbersToFiles = new Hashtable();
	private Hashtable		mNamesToFiles = new Hashtable();
	private Vector			mExclusionList = new Vector();
	private Vector			mInclusionList = new Vector();
	private Hashtable		mIndexedFiles = new Hashtable();
	private int				mTotalFileSize = 0;
	private Vector			mUploadList = new Vector();
	private int				mUploadingCount = 0;
	private IShareChanged	mShareChangedListener = null;
	private IndexerManager	mIndexerManager;
	private Hashtable		mMineTypeMapping = new Hashtable();
	private boolean			mRescanning = false;
	private DataChanger		mStatusChangedListener = new DataChanger();



	ShareManager()
	{
		mIndexerManager = ServiceManager.getIndexerManager();
		loadFileToContentTypeMappings();
	}


	void setShareChangedListener(IShareChanged listener)
	{
		mShareChangedListener = listener;
	}


	public void addStatusChangedListener(IDataChangedListener listener)
	{
		mStatusChangedListener.addListener(listener);
	}
	

	private void loadFileToContentTypeMappings()
	{
		mMineTypeMapping.put("txt", "text/plain");
		mMineTypeMapping.put("ini", "text/plain");
		mMineTypeMapping.put("c", "text/plain");
		mMineTypeMapping.put("h", "text/plain");
		mMineTypeMapping.put("cpp", "text/plain");
		mMineTypeMapping.put("cxx", "text/plain");
		mMineTypeMapping.put("cc", "text/plain");
		mMineTypeMapping.put("chh", "text/plain");
		mMineTypeMapping.put("java", "text/plain");
		mMineTypeMapping.put("csv", "text/plain");
		mMineTypeMapping.put("bat", "text/plain");
		mMineTypeMapping.put("cmd", "text/plain");
		mMineTypeMapping.put("sh", "text/plain");
		mMineTypeMapping.put("rtf", "text/richtext");
		mMineTypeMapping.put("rtx", "text/richtext");

		mMineTypeMapping.put("html", "text/html");
		mMineTypeMapping.put("htm", "text/html");
		
		mMineTypeMapping.put("zip", "application/zip");
		mMineTypeMapping.put("rar", "application/rar");
		mMineTypeMapping.put("gzip", "application/gzip");
		mMineTypeMapping.put("sit", "application/sit");
		mMineTypeMapping.put("tgz", "application/tgz");
		mMineTypeMapping.put("gz", "application/gz");
		mMineTypeMapping.put("tar", "application/tar");

		mMineTypeMapping.put("gif", "image/gif");
		mMineTypeMapping.put("jpeg", "image/jpeg");
		mMineTypeMapping.put("jpg", "image/jpeg");
		mMineTypeMapping.put("tiff", "image/tiff");
		mMineTypeMapping.put("tif", "image/tiff");
		mMineTypeMapping.put("png", "image/png");

		mMineTypeMapping.put("aiff", "audio/aiff");
		mMineTypeMapping.put("aif", "audio/aiff");
		mMineTypeMapping.put("au", "audio/basic");
		mMineTypeMapping.put("wav", "audio/wav");
		mMineTypeMapping.put("mp3", "audio/mp3");
		
		mMineTypeMapping.put("avi", "video/x-msvideo");
		mMineTypeMapping.put("mov", "video/quicktime");
		mMineTypeMapping.put("qt", "video/quicktime");
		mMineTypeMapping.put("mpeg", "video/mpeg");
		mMineTypeMapping.put("mpg", "video/mpeg");

		mMineTypeMapping.put("doc", "application/msword");
		mMineTypeMapping.put("xls", "application/msexcel");
		mMineTypeMapping.put("ps", "application/postscript");
		mMineTypeMapping.put("pdf", "application/pdf");
		
		mMineTypeMapping.put("exe", "application/binary");
		mMineTypeMapping.put("dll", "application/binary");
		mMineTypeMapping.put("class", "application/binary");
		mMineTypeMapping.put("jar", "application/binary");
	}


	public void setShareDirs(String dirs)
	{
		StringTokenizer	tokens = new StringTokenizer(dirs, ";");
		mDirsToShare.removeAllElements();

		while (tokens.hasMoreTokens())
		{
			String	dir = tokens.nextToken().trim();

			mDirsToShare.addElement(dir);
		}
	}


	public String getShareDirs()
	{
		String		str = "";

		for (int i = 0; i < mDirsToShare.size(); i++)
		{
			String	pattern = (String)mDirsToShare.elementAt(i);
			if (i > 0)
				str += ";";
			str += pattern;
		}

		return str;
	}


	public void addShareDir(String dir)
	{
		if (findSharedDir(dir) == -1)
			mDirsToShare.addElement(dir);
	}


	public void removeShareDir(String dir)
	{
		int	i = findSharedDir(dir);

		if (i != -1)
			removeShareDir(i);
	}


	public void removeShareDir(int i)
	{
		mDirsToShare.removeElementAt(i);
	}


	public void setExclusions(String str)
	{
		StringTokenizer	tokens = new StringTokenizer(str, ";");
		mExclusionList.removeAllElements();

		while (tokens.hasMoreTokens())
		{
			mExclusionList.addElement(tokens.nextToken().trim().toLowerCase());
		}
	}


	public String getExclusions()
	{
		String		str = "";

		for (int i = 0; i < mExclusionList.size(); i++)
		{
			String	pattern = (String)mExclusionList.elementAt(i);
			if (i > 0)
				str += ";";
			str += pattern;
		}

		return str;
	}	


	public void setInclusions(String str)
	{
		StringTokenizer	tokens = new StringTokenizer(str, ";");
		mInclusionList.removeAllElements();

		while (tokens.hasMoreTokens())
		{
			mInclusionList.addElement(tokens.nextToken().trim().toLowerCase());
		}
	}


	public String getInclusions()
	{
		String		str = "";

		for (int i = 0; i < mInclusionList.size(); i++)
		{
			String	pattern = (String)mInclusionList.elementAt(i);
			if (i > 0)
				str += ";";
			str += pattern;
		}

		return str;
	}


	public int getFileCount()
	{
		return mNumbersToFiles.size();
	}


	public ShareFile getFile(int index)
	{
		return (ShareFile)mFilesToShare.elementAt(index);
	}


	public ShareFile getFile(Integer fileNumber)
	{
		return (ShareFile)mNumbersToFiles.get(fileNumber);
	}


	public int getTotalFileSize()
	{
		return mTotalFileSize;
	}


	public Hashtable searchFile(String searchStr)
	{
		Vector			results = new Vector();
		Hashtable		resultsByNumber = new Hashtable();
		String[]		terms = new String[10];
		int				termCount = 0;
		StringTokenizer	tokens = new StringTokenizer(searchStr);

		// Build search term, upto 10 terms.
		while (termCount < 10 && tokens.hasMoreElements())
		{
			terms[termCount++] = tokens.nextToken();
		}

		// Don't search for empty string.
		if (termCount == 0)
			return groupFilesByHost(results);

		if (ServiceManager.sCfg.mPhexPingResponse && searchStr.equals("!`phex~~ping"))
		{
			ShareFile	shareFile = new ShareFile(null);
			results.addElement(shareFile);
			return groupFilesByHost(results);
		}

		// Don't response to any search if uploading limit has been reached.
		if (getUploadingCount() >= ServiceManager.sCfg.mMaxUpload)
			return groupFilesByHost(results);

		boolean			matchAnything = false;

		// Normalize search term.  Take care of wildcards.
		for (int i = 0; i < termCount; i++)
		{
			if (terms[i].equals("*") || terms[i].equals("*.*"))
			{
				matchAnything = true;
			}
			else
			{
				// Strip leading * and ending * for wildcard matches.
				if (terms[i].startsWith("*"))
				{
					terms[i] = terms[i].substring(1);
				}
				if (terms[i].endsWith("*"))
				{
					terms[i] = terms[i].substring(0, terms[i].length() - 1);
				}
			}
		}

		// Filename search.  Go through all the shared files and match the search terms to each filename.
		for (int i = 0; i < mFilesToShare.size(); i++)
		{
			ShareFile	shareFile = (ShareFile)mFilesToShare.elementAt(i);

			if (matchAnything || termsMatchName(terms, termCount, shareFile.getEffectiveName()))
			{
				if (resultsByNumber.get(shareFile.getFileIndex()) != null)
				{
					// Avoid duplicated result
					continue;
				}

				shareFile.incSearchCount(1);
				results.addElement(shareFile);
				resultsByNumber.put(shareFile.getFileIndex(), shareFile);
				if (results.size() >= ServiceManager.sCfg.mUploadMaxSearch)
				{
					// Result set size reaches max.  Don't search anymore.
					break;
				}
			}
		}

		// Content search.  Match the search terms to words of each file.
		Hashtable	filenumbers = termsMatchIndex(terms, termCount);
		if (filenumbers != null)
		{
			Enumeration	numbers = filenumbers.keys();
			while (numbers.hasMoreElements())
			{
				ShareFile	shareFile = (ShareFile)mNumbersToFiles.get(numbers.nextElement());

				if (resultsByNumber.get(shareFile.getFileIndex()) != null)
				{
					// Avoid duplicated result
					continue;
				}

				shareFile.incSearchCount(1);
				results.addElement(shareFile);
				resultsByNumber.put(shareFile.getFileIndex(), shareFile);
				if (results.size() >= ServiceManager.sCfg.mUploadMaxSearch)
				{
					// Result set size reaches max.  Don't search anymore.
					break;
				}
			}
		}

		return groupFilesByHost(results);
	}


	private Hashtable groupFilesByHost(Vector list)
	{
		// Collate files into groups.
		Hashtable	groups = new Hashtable();
		for (int i = 0; i < list.size(); i++)
		{
			ShareFile	sfile = (ShareFile)list.elementAt(i);
			Vector		group;

			if (sfile.getRemoteShareFileHost() == null)
			{
				group = (Vector)groups.get("localhost");
				if (group == null)
				{
					group = new Vector();
					groups.put("localhost", group);
				}
			}
			else
			{
				String	groupName = sfile.getRemoteShareFileHost().toString() + ":" + sfile.getRemoteShareFilePort();
				group = (Vector)groups.get(groupName);
				if (group == null)
				{
					group = new Vector();
					groups.put(groupName, group);
				}
			}
			group.addElement(sfile);
		}

		return groups;
	}


	private boolean termsMatchName(String[] terms, int termCount, String name)
	{
		// All terms have to be part of the name to be matched.
		for (int i = 0; i < termCount; i++)
		{
			if (name.indexOf(terms[i]) == -1)
			{
				return false;
			}
		}
		return true;
	}


	private Hashtable termsMatchIndex(String[] terms, int termCount)
	{
		Hashtable		result1;
		Hashtable		result2;

		result1 = (Hashtable)mIndexedFiles.get(terms[0].toUpperCase());
		if (result1 == null)
			return null;

		for (int i = 1; i < termCount; i++)
		{
			Hashtable	filenumbers = (Hashtable)mIndexedFiles.get(terms[i].toUpperCase());

			if (filenumbers == null)
				return null;

			result2 = new Hashtable();
			Enumeration	numbers = filenumbers.keys();
			while (numbers.hasMoreElements())
			{
				Integer		fileIndex = (Integer)numbers.nextElement();
				if (result1.get(fileIndex) != null)
				{
					result2.put(fileIndex, "");
				}
			}
			result1 = result2;
		}

		return result1;
	}


	public void rescan()
	{
		if (mRescanning)
		{
			return;
		}
		mRescanning = true;

		setShareDirs(ServiceManager.sCfg.mUploadDir);
		setExclusions(ServiceManager.sCfg.mUploadFileExclusions);
		setInclusions(ServiceManager.sCfg.mUploadFileInclusions);

		mFilesToShare.removeAllElements();
		mNumbersToFiles.clear();
		mNamesToFiles.clear();
		mTotalFileSize = 0;

		for (int i = 0; i < mDirsToShare.size(); i++)
		{
			String		dirName = (String)mDirsToShare.elementAt(i);
			File		dir = new File(dirName);

			scanDir(dir, ServiceManager.sCfg.mUploadScanRecursively);
		}
		shareChanged();

		mRescanning = false;
	}


	public void shareChanged()
	{
		if (mShareChangedListener != null)
			mShareChangedListener.shareChanged();
	}


	private void scanDir(File dir, boolean recursive)
	{
		// Resolve: wrote function for jdk 1.1.x, stepped down from 1.2.x.
//		File[]		files = dir.listFiles();
		File[]		files = getFileList(dir);

		if (files == null)
		{
			System.out.println("'" + dir + "' is not a directory.");
			return;
		}

		Properties	aliasMapping = loadAliasFile(dir);
		Properties	wordFiles = new Properties();

		if (!ServiceManager.sCfg.mIndexFiles)
		{
			try
			{
				FileInputStream	is = new FileInputStream(new File(dir, "files.idx"));
				wordFiles.load(is);
				is.close();
			}
			catch (IOException e)
			{
				// e.printStackTrace();
			}
		}

		for (int j = 0; j < files.length; j++)
		{
			if (skipFile(files[j]))
				continue;
			
			if (files[j].isFile())
			{
				if (ServiceManager.sCfg.mIndexFiles)
				{
					indexFile(wordFiles, files[j]);
				}

				ShareFile	shareFile = new ShareFile(files[j]);
				Integer		fileNumber = new Integer(Digest.foldBytesToInt(Digest.computeDigest(files[j].getAbsolutePath())));
				while (true)
				{
					// Make sure file number is unique.
					if (mNumbersToFiles.get(fileNumber) == null)
					{
						break;
					}
					fileNumber = new Integer(fileNumber.intValue() + 1);
				}
				shareFile.setFileIndex(fileNumber);
				shareFile.setAlias((String)aliasMapping.get(files[j].getName()));

				addWordIndex(files[j].getName(), fileNumber, wordFiles);

				mFilesToShare.addElement(shareFile);
				mNamesToFiles.put(files[j].getAbsolutePath(), fileNumber);
				mNumbersToFiles.put(fileNumber, shareFile);

				mTotalFileSize += files[j].length();

			}
		}

		if (ServiceManager.sCfg.mIndexFiles)
		{
			try
			{
				FileOutputStream	os = new FileOutputStream(new File(dir, "files.idx"));
				wordFiles.save(os, "PHEX's Words To Files Index");
				os.close();
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
		}

		if (recursive)
		{
			for (int j = 0; j < files.length; j++)
			{
				if (files[j].isDirectory())
				{
					scanDir(files[j], recursive);
				}
			}
		}

	}


	private File[] getFileList(File dir)
	{
		String[]	namelist = dir.list();
		Vector		tmplist = new Vector();
		String		parentPath = dir.getAbsolutePath();

		if (namelist == null)
			return null;
		
		for (int i = 0; i < namelist.length; i++)
		{
			String	filename = parentPath + File.separator + namelist[i];
			File	file = new File(filename);

			tmplist.addElement(file);
		}

		File[]		files = new File[tmplist.size()];
		for (int i = 0; i < tmplist.size(); i++)
		{
			files[i] = (File)tmplist.elementAt(i);
		}

		return files;
	}


	private int findSharedDir(String dir)
	{
		for (int i = 0; i < mDirsToShare.size(); i++)
		{
			String		dirName = (String)mDirsToShare.elementAt(i);

			if (dirName.equals(dir))
			{
				return i;
			}
		}

		return -1;
	}


	private void indexFile(Properties wordFiles, File file)
	{
		Hashtable		words;

		try
		{
			words = mIndexerManager.index(file, getFileExt(file.getName()));
		}
		catch (Exception e)
		{
			e.printStackTrace();
			return;
		}

		if (words == null)
			return;

		Enumeration		enum = words.keys();
		while (enum.hasMoreElements())
		{
			String		word = (String)enum.nextElement();
			String		filenames = (String)wordFiles.get(word);
			if (filenames == null)
			{
				filenames = "";
			}

			if (filenames.indexOf(file.getName() + ",") == -1)
				filenames += file.getName() + ",";
			wordFiles.put(word, filenames);
		}
	}


	private void addWordIndex(String filename, Integer fileNumber, Properties wordFiles)
	{
		Enumeration		keys = wordFiles.keys();

		while (keys.hasMoreElements())
		{
			String		word = (String)keys.nextElement();
			String		files = (String)wordFiles.get(word);
			if (files.indexOf(filename + ",") != -1)
			{
				Hashtable	filenumbers = (Hashtable)mIndexedFiles.get(word);

				if (filenumbers == null)
				{
					filenumbers = new Hashtable();
					mIndexedFiles.put(word, filenumbers);
				}

				filenumbers.put(fileNumber, "");
			}
		}
	}


	private String getFileExt(String filename)
	{
		int		index = filename.lastIndexOf(".");

		if (index > -1)
		{
			return filename.substring(index + 1, filename.length());
		}
		else
		{
			return "";
		}
	}


	private boolean skipFile(File file)
	{
		// In case the user is sharing the download directory,
		// skip the download-in-progress files.
		if (file.getName().toLowerCase().endsWith(".dl"))
			return true;
		// Skip PHEX generated index file.
		if (file.getName().equals("files.idx"))
			return true;
		// Skip alias file.
		if (file.getName().equals(sAliasFilename))
			return true;

		if (!isIncluded(file))
			return true;

		if (isExcluded(file))
			return true;

		return false;
	}


	private boolean isIncluded(File file)
	{
		String		name = file.getName().toLowerCase();

		for (int i = 0; i < mInclusionList.size(); i++)
		{
			String	pattern = (String)mInclusionList.elementAt(i);

			if (pattern.equals("*"))
			{
				return true;
			}
			if (pattern.equals("*.*"))
			{
				return true;
			}

			// Take care of *exp and exp* cases.
			if (pattern.startsWith("*"))
			{
				pattern = pattern.substring(1);
			}
			if (pattern.endsWith("*"))
			{
				pattern = pattern.substring(0, pattern.length() - 1);
			}

			if (name.indexOf(pattern) != -1)
			{
				return true;
			}
		}

		return false;
	}


	private boolean isExcluded(File file)
	{
		String		name = file.getName().toLowerCase();

		// Exclude any file contain the "filealias" string.
		if (name.indexOf(sAliasFile) != -1)
			return true;

		for (int i = 0; i < mExclusionList.size(); i++)
		{
			String	pattern = (String)mExclusionList.elementAt(i);

			if (name.indexOf(pattern) != -1)
			{
				return true;
			}
		}

		return false;
	}


	private Properties loadAliasFile(File parent)
	{
		Properties	aliases = new Properties();
		try
		{
			FileInputStream	is = new FileInputStream(new File(parent, sAliasFilename));
			aliases.load(is);
			is.close();
		}
		catch (Exception e)
		{
		}
		return aliases;
	}


	public Vector getUploadList()
	{
		return mUploadList;
	}


	public synchronized void addUploadFile(UploadFile file)
	{
		mUploadList.addElement(file);
	}


	public synchronized void removeUploadFile(int index)
	{
		if (index >= mUploadList.size())
			return;

		mUploadList.removeElementAt(index);
	}


	public synchronized void abortUpload(int index)
	{
		if (index >= mUploadList.size())
			return;

		UploadFile	file = (UploadFile)mUploadList.elementAt(index);
		Host		host = file.getRemoteHost();

		file.setStatus(UploadFile.sAborted);
		host.closeConnection();

	}


	public synchronized void clearUpload()
	{
		for (int i = 0; i < mUploadList.size(); )
		{
			UploadFile	file = (UploadFile)mUploadList.elementAt(i);

			if (file.getStatus() == UploadFile.sQueued || file.getStatus() == UploadFile.sUploading)
			{
				i++;
				continue;
			}

			mUploadList.removeElementAt(i);
		}

	}


	public synchronized int getUploadingCount()
	{
		return mUploadingCount;
	}


	public synchronized void incUploadingCount()
	{
		mUploadingCount++;
		mStatusChangedListener.dataChanged(null);
	}


	public synchronized void decUploadingCount()
	{
		mUploadingCount--;
		mStatusChangedListener.dataChanged(null);
	}


	private int getUploadingPerIPCount(Host remoteHost)
	{
		int		count = 0;

		for (int i = 0; i < mUploadList.size(); i++)
		{
			UploadFile	file = (UploadFile)mUploadList.elementAt(i);
			Host		host = file.getRemoteHost();
			if (file.getStatus() == UploadFile.sUploading &&
				host.getHostAddr().equals(host.getHostAddr()))
			{
				count++;
			}
		}

		return count;
	}



	// Called by ReadWorker to handle a HTTP GET request from the remote host.
	public void httpRequestHandler(ConnectionRequest req, Host remoteHost)
	{
		if (!req.readHeaders(remoteHost.getIs()))
		{
			// Error in reading HTTP readers.  Just return and close connection.
			return;
		}

		// GET / HTTP/1.1
		if (req.getUri().equals("/"))
		{
			// The remote host just want the index.html.
			// Return a list of shared files.
			sendFile(req, remoteHost);
			return;
		}

		// GET /get/1/foo.txt HTTP/1.1
		StringTokenizer	tokens = new StringTokenizer(req.getUri(), "/");
		String			operation;
		int				fileIndex;
		String			fileName;
		try
		{
			// /get/1/foo.txt
			operation = tokens.nextToken();
			fileIndex = Integer.parseInt(tokens.nextToken());
			fileName = tokens.nextToken();
		}
		catch (Exception e)
		{
			// Didn't get all the expected fields.  Wrong request.
			// Just return a list of shared files.
			sendFile(req, remoteHost);
			return;
		}

		if (!operation.toLowerCase().equals("get"))
		{
			// Wrong operation.
			sendFile(req, remoteHost);
			return;
		}

		ShareFile	shareFile = (ShareFile)mNumbersToFiles.get(new Integer(fileIndex));
		if (shareFile == null)
		{
			// Can't file file.
			sendString(remoteHost, buildErrorHTTP("404 Not Found", "File not found"));
			return;
		}

		// Finally things look ok.  Go ahead and do upload.
		File		file = shareFile.getFile();

		if (upload(req, remoteHost, file, true))
			shareFile.incUploadCount(1);
	}


//	private void upload(ConnectionRequest req, Host remoteHost, int fileIndex, String fileName, boolean gnutellaUpload)
	private boolean upload(ConnectionRequest req, Host remoteHost, File file, boolean gnutellaUpload)
	{			
		RandomAccessFile	fis = null;
		UploadFile			localfile = null;
		HostManager			hm = ServiceManager.getHostManager();
		boolean				rc = false;

		try
		{
			// Check if allow further upload
			if (!ServiceManager.sCfg.mShareEnabled)
			{
				remoteHost.log("Upload request: upload disabled.  No more upload allowed.");
				sendString(remoteHost, buildErrorHTTP("404 Not Found", "File sharing disabled"));
				return rc;
			}

			// Control the number of simultaneous uploads.
			if (getUploadingCount() >= ServiceManager.sCfg.mMaxUpload)
			{
				remoteHost.log("Upload request: upload limit reached.");
				sendString(remoteHost, buildErrorHTTP("503 Service Unavailable", "Server busy.  Too many downloads from this host."));
				return rc;
			}

			// Control the number of simultaneous uploads per IP.
			if (getUploadingPerIPCount(remoteHost) >= ServiceManager.sCfg.mMaxUploadPerIP)
			{
				remoteHost.log("Upload request: upload limit per host reached.");
				sendString(remoteHost, buildErrorHTTP("503 Service Unavailable", "You have too many downloads from this host."));
				return rc;
			}

			// See if it's a partial get.
			HttpRange	range = new HttpRange((int)file.length());
			String		rangeValue = req.getHeader("RANGE");
			if (rangeValue != null)
			{
				if (!range.deserialize(rangeValue))
				{
					// Invalide range value.
					remoteHost.log("Upload request: Invalid range value in partial get.  " + rangeValue);
					sendString(remoteHost, buildErrorHTTP("416 Requested Range Not Satisfiable", "Upload request: Invalid range value in partial get.  " + rangeValue));
					return rc;
				}
				if (range.getSpecCount() == 0)
				{
					range.addSpec(0, (int)file.length() - 1);
				}
			}
			else
			{
				range.addSpec(0, (int)file.length() - 1);
			}

			incUploadingCount();

			int				transferLen = range.getLastPos() - range.getFirstPos() + 1;

			localfile = new UploadFile(file, remoteHost, transferLen, req.getHeader("USER-AGENT"));
			localfile.setStatus(UploadFile.sUploading);
			localfile.setStartingTime(System.currentTimeMillis());
			addUploadFile(localfile);

			// Prepare HTTP response
			byte[]			outbuf = new byte[512];
			String			headers = "";
			String			contentType = "application/binary";

			if (!gnutellaUpload)
			{
				// Mapping file type to content type.
				String		ctype = (String)mMineTypeMapping.get(StrUtil.getFileExt(localfile.getFilename()).toLowerCase());
				if (ctype != null)
					contentType = ctype;
			}


			if (transferLen == range.getEntitySize())
			{
				headers += "HTTP/1.1 200 OK\r\n";
			}
			else
			{
				if (gnutellaUpload)
				{
					// Gnutella client expects non-standard syntax for partial GET.
					headers += "HTTP/1.1 200 OK\r\n";
				}
				else
				{
					// Use standard HTTP syntax.
					headers += "HTTP/1.1 206 OK\r\n";
				}
			}
			headers += "Server: " + StrUtil.getAppNameVersion() + "\r\n";
			headers += "Accept-Ranges: bytes\r\n";
			headers += "Connection: close\r\n";
			headers += "Content-Type: " + contentType + "\r\n";
			if (transferLen == range.getEntitySize())
			{
				headers += "Content-length: " + range.getEntitySize() + "\r\n";
			}
			else
			{
				if (gnutellaUpload)
				{
					// Gnutella client expects non-standard syntax for Content-range.
					headers += "Content-range: bytes=" + range.getFirstPos() + "-" + range.getLastPos() + "/" + range.getEntitySize() + "\r\n";
				}
				else
				{
					// Use standard HTTP syntax.
					headers += "Content-range: bytes " + range.getFirstPos() + "-" + range.getLastPos() + "/" + range.getEntitySize() + "\r\n";
					headers += "Content-length: " + transferLen + "\r\n";
				}
			}
			headers += "\r\n";

			// Write HTTP response headers
			int				len = IOUtil.serializeString(headers, outbuf, 0);
			OutputStream	os = remoteHost.getOs();
			os.write(outbuf, 0, len);

			fis = new RandomAccessFile(file, "r");
			fis.seek(range.getFirstPos());

			long			count = 0;
			int				sleep = 1;
			while (transferLen > 0)
			{
				len = fis.read(outbuf);
				if (len == -1)
					break;
				os.write(outbuf, 0, len);
				transferLen -= len;
				count += len;
				localfile.setTransferredSize(count);

				// Throttle control
				if (localfile.getTransferRate() > ServiceManager.sCfg.mUploadMaxBandwidth / mUploadingCount)
				{
					sleep = sleep * 2;
				}
				else if (localfile.getTransferRate() < ServiceManager.sCfg.mUploadMaxBandwidth / mUploadingCount)
				{
					sleep = sleep / 2;
					if (sleep == 0)
						sleep = 1;
				}
				Thread.sleep(sleep);
			}

			localfile.setTransferredSize(count);
			localfile.setStatus(UploadFile.sCompleted);
			hm.incStatUploadCount(1);
			rc = true;

			// Wait a bit before closing the connection.
			// Somehow the remote gnutella won't read the last
			// buffer if closing the connection too soon.
			Thread.sleep(2000);
		}
		catch (Exception e)
		{
			rc = false;
//			e.printStackTrace();
			if (localfile != null && localfile.getStatus() != UploadFile.sAborted)
			{
				localfile.setStatus(UploadFile.sError, e.getMessage());
			}
		}

		if (fis != null)
		{
			try
			{
				fis.close();
			}
			catch (IOException e)
			{
			}
		}

		decUploadingCount();

		return rc;
	}


	private void sendFile(ConnectionRequest req, Host remoteHost)
	{
		String			uri;

		try
		{
			uri = URLUtil.decodeURL(req.getUri());
		}
		catch (Exception e)
		{
			remoteHost.log("Invalid URI from client.  " + req.getUri());
			sendString(remoteHost, buildErrorHTTP("400 Bad Request", "Invalid URI from client.  " + req.getUri()));
			return;
		}

		if (uri.indexOf("..") != -1)
		{
			remoteHost.log("Invalid URI from client, containing '..'  " + req.getUri());
			sendString(remoteHost, buildErrorHTTP("400 Bad Request", "Invalid URI from client.  " + req.getUri()));
			return;
		}

		for (int i = 0; i < mDirsToShare.size(); i++)
		{
			String		dirName = (String)mDirsToShare.elementAt(i);
			File		dir = new File(dirName);
			String		normalizedUri = StrUtil.replaceAllChar(uri, '/', File.separator);
			String		path = dir.getAbsolutePath() + normalizedUri;
			File		file = new File(path);


			if (!file.exists())
				continue;

			if (!file.isFile())
			{
				File	defaultFile;

				defaultFile = new File(file, "index.html");
				if (defaultFile.exists())
				{
					upload(req, remoteHost, defaultFile, false);
					return;
				}
				defaultFile = new File(file, "Index.html");
				if (defaultFile.exists())
				{
					upload(req, remoteHost, defaultFile, false);
					return;
				}
				defaultFile = new File(file, "default.html");
				if (defaultFile.exists())
				{
					upload(req, remoteHost, defaultFile, false);
					return;
				}
				defaultFile = new File(file, "Default.html");
				if (defaultFile.exists())
				{
					upload(req, remoteHost, defaultFile, false);
					return;
				}

				if (!ServiceManager.sCfg.mShareBrowseDir)
				{
					sendString(remoteHost, buildErrorHTTP("403 Forbidden", "Directory browsing is not allowed."));
				}
				else
				{
					sendString(remoteHost, buildDirListHTTP(file, uri));
				}
				return;
			}

			Integer		fileNumber = (Integer)mNamesToFiles.get(file.getAbsolutePath());

			if (fileNumber != null)
			{
				ShareFile	shareFile = (ShareFile)mNumbersToFiles.get(fileNumber);
				File		file2 = shareFile.getFile();

				if (upload(req, remoteHost, file2, false))
					shareFile.incUploadCount(1);
			}
			else
			{
				sendString(remoteHost, buildErrorHTTP("404 Not Found", "File not found."));
			}
			return;
		}

		sendString(remoteHost, buildErrorHTTP("404 Not Found", "File not found."));
	}


	private void sendString(Host remoteHost, String html)
	{
		try
		{
			// Prepare HTTP response
			byte[]			outbuf = new byte[html.length()];
			int				lenToSend = IOUtil.serializeString(html, outbuf, 0);
			int				lenSent = 0;
			int				len;
			OutputStream	os = remoteHost.getOs();
			HostManager		hm = ServiceManager.getHostManager();
			int				sleep = 1;

			// Write HTTP response
			while (lenSent < lenToSend)
			{
				len = lenToSend - lenSent;
				if (len > 1024)
					len = 1024;
				os.write(outbuf, lenSent, len);
				lenSent += len;

				sleep = hm.throttleControl(sleep, 1000, hm.getRate(), ServiceManager.sCfg.mNetMaxRate);
			}

			// Wait a bit before closing the connection.
			// Somehow the remote gnutella won't read the last
			// buffer if closing the connection too soon.
			Thread.sleep(2000);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}


	private String buildDirListHTTP(File dir, String uri)
	{
		StringBuffer		buf = new StringBuffer();
		StringBuffer		body = new StringBuffer();
		File[]				files;
		String				field;
		String				date;
		String				size;
		SimpleDateFormat	dformat = new SimpleDateFormat("yyyy/MM/dd KK:mma");


		// Build the HTTP header
		buf.append("HTTP/1.1 200 OK\r\n");
		buf.append("Server: " + StrUtil.getAppNameVersion() + "\r\n");
		buf.append("Content-Type: text/html\r\n");
		buf.append("Connection: close\r\n");

		// Build the HTTP body
		body.append("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">");
		body.append("<HTML>");
		body.append("<HEAD><TITLE>Index of " + uri + "</TITLE></HEAD>");
		body.append("<BODY>");

		body.append("<H3>Index of " + uri + "</H3>\n");
		body.append("<PRE>");

		body.append("Name                                   Last Modified           Size       Bytes\n");
		body.append("<HR>");

		if (uri.startsWith("/") && uri.length() > 1)
		{
			// A subdirectory.
			// Build link to parent directory.
			if (uri.endsWith("/"))
				uri = uri.substring(0, uri.length() - 1);
			
			int	index = uri.lastIndexOf("/");
			uri = uri.substring(0, index + 1);

			String	encStr = URLUtil.encodeURL(uri);
			body.append("[<A HREF=\"" + encStr + "\">Parent Directory</A>]\n");

			// Get a list of files in the subdirectory.
			files = getFileList(dir);
		}
		else
		{
			// Root directory.  Get a list of files of all the top directories.
			Vector			list = new Vector();

			for (int i = 0; i < mDirsToShare.size(); i++)
			{
				String		dirName = (String)mDirsToShare.elementAt(i);
				files = getFileList(new File(dirName));
				for (int j = 0; j < files.length; j++)
				{
					list.addElement(files[j]);
				}
			}

			files = new File[list.size()];
			for (int i = 0; i < list.size(); i++)
			{
				files[i] = (File)list.elementAt(i);
			}
		}

		// Output the file list to a HTML page.
		for (int j = 0; j < files.length; j++)
		{
			if (skipFile(files[j]))
				continue;

			String			filename = files[j].getName();

			if (!files[j].isFile())
			{
				filename += "/";
			}
			
			String			encStr = URLUtil.encodeURL(filename);
			String			paddedFilename = StrUtil.getPadStr(filename, 39);
			StringBuffer	filenameHTML = new StringBuffer(paddedFilename);
			StrUtil.replaceAllChar(filenameHTML, '&', "&amp;");
			StrUtil.replaceAllChar(filenameHTML, '\"', "&quot;");
			StrUtil.replaceAllChar(filenameHTML, '<', "&lt;");
			StrUtil.replaceAllChar(filenameHTML, '>', "&gt;");

			field = "<A HREF=\"" + encStr + "\">" + filename + "</A>" + filenameHTML;
			body.append(field);

			date = dformat.format(new java.util.Date(files[j].lastModified()));
			body.append(StrUtil.padStringRight(date, 19));

			size = (files[j].isFile() ? StrUtil.formatSizeBytes(files[j].length()) : "-");
			body.append(StrUtil.padStringLeft(size, 9));

			size = (files[j].isFile() ? String.valueOf(files[j].length()) : "-");
			body.append(StrUtil.padStringLeft(size, 12));

			body.append("\n");
		}

		body.append("<HR>");
		body.append("Powered by " + StrUtil.getAppNameVersion() +
					"\r\n<A HREF=\"http://www.jps.net/williamw/phex/\">http://www.jps.net/williamw/phex/</A>\r\n");
		body.append("</PRE>");
		body.append("</BODY>");
		body.append("</HTML>");

		// Finish up the header and add in the body.
		buf.append("Content-Length: " + body.length() + "\r\n");
		buf.append("\r\n");
		buf.append(body);

		return buf.toString();
	}


	private String buildErrorHTTP(String statusStr, String errMsg)
	{
		StringBuffer		buf = new StringBuffer();

		buf.append("HTTP/1.1 " + statusStr + "\r\n");
		buf.append("Server: " + StrUtil.getAppNameVersion() + "\r\n");
		buf.append("Connection: close\r\n");
		buf.append("Content-Tyep: text/plain\r\n");
		buf.append("Content-Length: " + errMsg.length() + "\r\n");
		buf.append("\r\n");
		buf.append(errMsg);

		return buf.toString();
	}


}


